/* $Workfile: A66Slic2.c $       $Revision: 39 $      $Modtime: 5/22/08 12:04p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    A66Slic2.c

    A basic example console application demonstrating the use of OTX Driver
    with the Arni-6x6-PCI or the Arni-6x6-PCI board.

    The sample application configures 
    6 FXS SLIC Line Interfaces. 
    
    The application provides
    a simple console interface to manage the SLIC Li devices.

    The user can use the following simple keystrokes as commands to go off-hook
    and on-hook, to dial numbers, etc.

    In this basic demo the FXS SLIC LIs are supposed to be connected
    to telephones. The program does not demonstrate the advanced functions of the 
    Arni board, such a prompt playing, conferencing, etc. For such functions please
    refer to more advanced Arni demo programs.
    
    The main command keys are listed in the following:
     - Press 'h' or '?' for help. 'q' to quit.
     

*******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "OtxSlic.h"
#include "OtxData.h"

#include "OtxBurst.h"
#include "OtxBurstD.h"


/*******************************************************************************
  Macros, Types
*******************************************************************************/

//#define DEMO_DSP_MAX 6
#define DEMO_LI_MAX  6

#define DEMO_TASK_REF           1L   // Task Reference for a non-blocking call
#define DEMO_DEFAULT_LOG_FILE   "A66Slic2.log"       // Name of the LOG file
#define DEMO_DEFAULT_BOARD_NO   0
#define CURRENT_LI              0 // Li that is used as primary
#define CONNECTION_LI           1 // LI that is called by the primary
#define DEMO_NO_LI_CONNECTED   0xFF

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k
// U-law samples for playback over the FXS device 
// The program plays this file from disk.
#if defined(__linux__)
#define DEMO_FXS_INPUT_FILE   "..//..//..//..//PromptFileUlaw.snd" 
#elif defined(WIN32)
#ifdef DEMO_FXS_INPUT_FILE
#undef DEMO_FXS_INPUT_FILE
#endif 
#define DEMO_FXS_INPUT_FILE   "..//PromptFileUlaw.snd" 
#endif   
#define DEMO_U_LAW_SILENCE   0xFF    // Sending this byte produces silence in u-law mode

typedef struct DemoEnvT {
  char                     *pszInputStr;
  OTX_BOOL                  bExitOnError;
  char                     *logFileName;
  int                       boardNo;  
  OTX_INT32                 diagTest;
  OtxPhysicalDeviceTypeE    nBoardType;
} DemoEnvT;


struct {  // Handles to Pipes
  OTX_HANDLE hPotsRx[DEMO_LI_MAX];
  OTX_HANDLE hPotsTx[DEMO_LI_MAX];
} g_hPipes;




/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConnectLiToLiDevices( void );
static void DemoDisconnectLiDevices( void );
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void *DemoSlicEventFunction( void *lpParameter  /* thread data */ ); 
static void stateMachine(OTX_UINT32 deviceNumber);
static void slicLoopTestOffHook(OTX_UINT32 deviceNumber);
static void slicLoopTestOnHook(OTX_UINT32 deviceNumber);
static void slicLoopTest(OTX_UINT32 deviceNumber);
static void setState(OTX_UINT32 newState, OTX_UINT32 deviceNumber);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCallLineMakeRing(void);
static void printPhoneNumbers(void);
static void dtmfAction(OTX_UINT8 rawDigit, OTX_UINT32 deviceNumber);
static void DemoBurstSender(OTX_UINT32 nSourceId);
static void DemoBurstReceiver(OTX_UINT32 nSourceId);
static OTX_BOOL DemoIsBurstDeviceEnabled(void);
static void setupCaptureLiData(void);
static void ringAllLines(void);
static OTX_INT32 DemoProcessUserCommand(char commandChar);
static OTX_INT32 DemoSlicReceiveConfigMenu(char commandChar);



#define NUMBER_OF_DTMF_DIGITS 20 
                          
/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue        = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE  g_hPipeSendSlic      = OTX_INVALID_HANDLE_VALUE;
FILE       *g_pDemoLogFile       = NULL;
DemoEnvT    g_demoEnv;
// Demo needs two Si3210 devices to make a connection
OTX_UINT32  g_nCurrentLi         = CURRENT_LI;      // Li POTS 1
OTX_UINT32  g_dtmfDeviceNumber;
// This example will support a single set of connection at one time
OTX_UINT32  g_liConDeviceNumber_1 = DEMO_NO_LI_CONNECTED;
OTX_UINT32  g_liConDeviceNumber_2 = DEMO_NO_LI_CONNECTED;
OTX_UINT32  g_RunNormalMode       = OTX_TRUE;


OTX_UINT8   g_rawDigit;
OTX_BOOL    g_bWatchDogThreadActive = OTX_FALSE;
OTX_BOOL    g_bSendCallerId = OTX_FALSE;

OTX_BOOL    g_bBurstEnabled = OTX_FALSE;
OTX_BOOL    g_bLibConnected = OTX_FALSE;
OTX_BOOL    g_bDemoExitQuestion = OTX_TRUE; // Disable during active call

// Global variables used to store call state for each slic
static OTX_UINT32 chipDataRingCount[DEMO_LI_MAX];
static OTX_UINT32 chipDataDigitCount[DEMO_LI_MAX];
static OTX_UINT32 chipDataState[DEMO_LI_MAX];
static OTX_UINT32 chipDataNewState[DEMO_LI_MAX];
static OTX_UINT32 chipDataPreviousState[DEMO_LI_MAX];
static OTX_UINT8  chipDataDtmfDigits[DEMO_LI_MAX][NUMBER_OF_DTMF_DIGITS];
static OsaMutexS g_demoLogfCritSec;

struct {  // Handles to Physical Devices
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hSlicLi[6];
  OTX_HANDLE hBurst;
  OTX_HANDLE hTss;
} g_hPhysDevs;

// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  OTX_UINT8   sqBufCurr[DEMO_BURST_BUF_SIZE / 32]; // For file reads
  OTX_UINT8   sqBufPrev[DEMO_BURST_BUF_SIZE / 32]; // For file reads (the previous buffer)
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  
};

struct {
  struct sBurstSender mBurstSender;        
  struct sBurstReceiver mBurstReceiver;    
} g_sBurstDevs;


// Call states for Slic device
enum DemoSlicDeviceCallState { 
  MAKE_BUSY                   = OTX_SLIC_CS_END_OF_STATES + 0 , 
  STATE_CHANGE                = OTX_SLIC_CS_END_OF_STATES + 1,
  INITIALIZING                = OTX_SLIC_CS_END_OF_STATES + 2,
  MAKE_RING_BACK              = OTX_SLIC_CS_END_OF_STATES + 3,
  MAKE_REORDER                = OTX_SLIC_CS_END_OF_STATES + 4, 
  MAKE_CONGESTION             = OTX_SLIC_CS_END_OF_STATES + 5,
  CALLBACK_PROGRAMMED         = OTX_SLIC_CS_END_OF_STATES + 6,
  CALL_BACK                   = OTX_SLIC_CS_END_OF_STATES + 7,
  MAKE_OFFHOOK                = OTX_SLIC_CS_END_OF_STATES + 8,
  ONHOOK                      = OTX_SLIC_CS_END_OF_STATES + 9,
  OFFHOOK                     = OTX_SLIC_CS_END_OF_STATES + 10,
  FLASHHOOK                   = OTX_SLIC_CS_END_OF_STATES + 11,
  DIGIT_DECODING              = OTX_SLIC_CS_END_OF_STATES + 12,
  LOOP_TRANSITION             = OTX_SLIC_CS_END_OF_STATES + 13, 
  CALLER_ID                   = OTX_SLIC_CS_END_OF_STATES + 14,
  RINGING                     = OTX_SLIC_CS_END_OF_STATES + 15,
  STOP_RINGING                = OTX_SLIC_CS_END_OF_STATES + 16,
  DTMF_TRANISTION             = OTX_SLIC_CS_END_OF_STATES + 17,
  NO_ACTION_NEEDED            = OTX_SLIC_CS_END_OF_STATES + 18
};
typedef enum DemoSlicDeviceCallState DemoSlicDeviceCallState;


// Call states for Slic device
char *g_sqszSlicDeviceCallState[] = { 
  // OTX Driver states
  "(drv)INITIALIZING                "    ,
  "(drv)CALIBRATE                   "    ,
  "(drv)MAKE_OFFHOOK                "    ,
  "(drv)ONHOOK                      "    ,
  "(drv)OFFHOOK                     "    ,
  "(drv)FIRST_RING                  "    ,
  "(drv)DEFERED_CALLBACK_PROGRAMMED "    ,
  "(drv)CALLER_ID                   "    ,
  "(drv)RINGING                     "    ,
  "(drv)DTMF_TRANSITION             "    ,
                                    
   // Demo App states               
  "(app)MAKE_BUSY                   "    ,
  "(app)STATE_CHANGE                "    ,
  "(app)INITIALIZING                "    ,
  "(app)MAKE_RING_BACK              "    ,
  "(app)MAKE_REORDER                "    ,
  "(app)MAKE_CONGESTION             "    ,
  "(app)CALLBACK_PROGRAMMED         "    ,
  "(app)CALL_BACK                   "    ,
  "(app)MAKE_OFFHOOK                "    ,
  "(app)ONHOOK                      "    ,
  "(app)OFFHOOK                     "    ,
  "(app)FLASHHOOK                   "    ,
  "(app)DIGIT_DECODING              "    ,
  "(app)LOOP_TRANSITION             "    ,
  "(app)CALLER_ID                   "    ,
  "(app)RINGING                     "    ,
  "(app)STOP_RINGING                "    ,
  "(app)DTMF_TRANISTION             "    ,
  "(app)NO_ACTION_NEEDED            "    ,
                                   
};



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); 
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); 
  DemoCheckRc(g_hEventQueue, result);

}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Opens the Physical Devices on the Arni-16-PCI board

  @comm
    Calls function <f OtxDrvOpenPhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Get Handle to the Physical ARNI Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_ARNI_6X6_PCI_PLUS:
    DemoLogf("Opening Arni-6x6-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6_FXS_PCI_PLUS:
    DemoLogf("Opening Arni-6-FXS-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X3_PCI_PLUS:
    DemoLogf("Opening Arni-6x3-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X6_PCI:
    DemoLogf("Opening Arni-6x6-PCI board\n");
    break;
  case OTX_DEVICE_ARNI_6_FXO_PCI_PLUS:
    DemoLogf("Opening Arni-6-FXO-PCI-Plus board\n");
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", 
             g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, 
                                    g_demoEnv.nBoardType, 
                                    g_demoEnv.boardNo, 
                                    g_demoEnv.boardNo, 
                                    g_hEventQueue, 
                                    &g_hPhysDevs.hBoardController0); 
  DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Check if the watchdog timer has triggered before we continue
  {
    OTX_UINT32 readVal;
    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0, 0x04, &readVal); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (readVal == 0) {
      DemoLogf("The watchdog timer has triggered. A reboot of the PC is necessary.\n");
      DemoExitQuestion();
    }
  }

  // Disable the Watchdog Timer, and enable it when we have completed the configuration
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, 
                                       OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  
                                       OTX_FALSE); 
  DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0 ; i < DEMO_LI_MAX; i++) {
    DemoLogf("Open OTX_DEVICE_LI_SLIC  #%d \n", i);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, 
                                      OTX_DEVICE_LI_SLIC, 
                                      i, 
                                      i, 
                                      g_hEventQueue, 
                                      &g_hPhysDevs.hSlicLi[i]); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);
    
    // Take the fail-over relays out of the pass-through state 
    // (i.e. engage the SLIC on the line)
    DemoLogf("Take the fail-over relay out of "
             "the pass-through state for SLIC #%d\n", i);
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hSlicLi[i], 
                                         OTX_ATTR_SLIC_RELAY_CONTROL, 
                                         OTX_TRUE); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);    
  }

  // Get a handle to the Time-Space Switch (TSS) on the board
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, 
                                    OTX_DEVICE_TSS, 
                                    0, 
                                    0, 
                                    g_hEventQueue, 
                                    &g_hPhysDevs.hTss); 
  DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvEnable(g_hPhysDevs.hTss, OTX_TASK_SYNC, 5000L); 
  DemoCheckRc(g_hPhysDevs.hTss, result);

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoEnablePhysicalDevices\n");
  result = OtxDrvEnable(g_hPhysDevs.hBoardController0, OTX_TASK_SYNC, 5000L); 
  DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0; i < DEMO_LI_MAX; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hSlicLi[i], OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);
  }

}


static void DemoConnectLiToLiDevices( void )
/*******************************************************************************
  @func
    Create the needed Pipe Devices and connect the physical and 
    logical devices to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream 
    (a subset of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. 
    A pipe is a characterized by its capacity (in 8 kbit/s increments). 
    The user can create a pipe with the driver function <f OtxDrvCreatePipe>.
    A pipe is connected to physical or logical devices and mapped to
    specific time-slots with the driver function <f OtxDrvConnectPipe>.
    
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 slic;

  // This Demo connect two devices together
  DemoLogf("DemoConnectLiToLiDevices\n");

  // Data from and to POTS LIs
  if( (DEMO_NO_LI_CONNECTED == g_liConDeviceNumber_1) || 
      (DEMO_NO_LI_CONNECTED == g_liConDeviceNumber_2) )  {
    // The demo has a sequence issue no connects are set
    DemoLogf("Warning DemoConnectLiToLiDevices: No connections were set!\n");
  }
  else {
    slic = g_liConDeviceNumber_1;
    result = OtxDrvCreatePipe(64, &g_hPipes.hPotsRx[slic]); 
    DemoCheckRc(g_hPipes.hPotsRx[slic], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hPotsTx[slic]); 
    DemoCheckRc(g_hPipes.hPotsTx[slic], result);
  
    // Connect Pipes
    // Connect device 1 to the device 2 as an input
    result = OtxDrvConnectPipe(g_hPipes.hPotsRx[slic], 
                               OTX_PIPE_INPUT,  
                               g_hPhysDevs.hSlicLi[g_liConDeviceNumber_1], 
                               6,                 // 6 is Slic Physical Highway
                               0,                 // sqTimeSlotMask
                               0);                // Size of connection mask
    DemoCheckRc(g_hPipes.hPotsRx[slic], result);
    // Connect device 1 to the device 2 as an output
    result = OtxDrvConnectPipe(g_hPipes.hPotsRx[slic], 
                               OTX_PIPE_OUTPUT,  
                               g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2], 
                               6,                 // 6 is Slic Physical Highway
                               0,                 // sqTimeSlotMask
                               0);                // Size of connection mask
    DemoCheckRc(g_hPipes.hPotsRx[slic], result);

    // Connect device 1 to the device 2 as an output
    result = OtxDrvConnectPipe(g_hPipes.hPotsTx[slic],
                               OTX_PIPE_OUTPUT,  
                               g_hPhysDevs.hSlicLi[g_liConDeviceNumber_1], 
                               6,                 // 6 is Slic Physical Highway
                               0,                 // sqTimeSlotMask
                               0);                // Size of connection mask
    // Connect device 1 to the device 2 as an input
    result = OtxDrvConnectPipe(g_hPipes.hPotsTx[slic],
                               OTX_PIPE_INPUT,  
                               g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2], 
                               6,                 // 6 is Slic Physical Highway
                               0,                 // sqTimeSlotMask
                               0);                // Size of connection mask
    DemoCheckRc(g_hPipes.hPotsTx[slic], result);
  }

  DemoLogf("Done\n");

  return;
}


void DemoDisconnectLiDevices( void )
/*******************************************************************************
  @func
    Disconnect a pipe

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream 
    (a subset of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. 
    A pipe is a characterized by its capacity (in 8 kbit/s increments). 
    The user can create a pipe with the driver function <f OtxDrvCreatePipe>.
    A pipe is connected to physical or logical devices and mapped to
    specific time-slots with the driver function <f OtxDrvConnectPipe>.
    
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 slic;

  // This Demo connected two devices together
  DemoLogf("DemoDisconnectLiDevices\n");

  if( (DEMO_NO_LI_CONNECTED == g_liConDeviceNumber_1) || 
      (DEMO_NO_LI_CONNECTED == g_liConDeviceNumber_2) )  {
    DemoLogf("   No Li were connected\n");
  }

  // Data from and to POTS LIs
  slic = g_liConDeviceNumber_1;
  if(g_liConDeviceNumber_1 != DEMO_NO_LI_CONNECTED) {
    result = OtxDrvDisconnectPipe(g_hPipes.hPotsTx[slic], OTX_PIPE_OUTPUT); 
    DemoCheckRc(g_hPipes.hPotsTx[slic], result);
    result = OtxDrvDisconnectPipe(g_hPipes.hPotsRx[slic], OTX_PIPE_INPUT); 
    DemoCheckRc(g_hPipes.hPotsRx[slic], result);
  }
  if(g_liConDeviceNumber_2 != DEMO_NO_LI_CONNECTED) {
    result = OtxDrvDisconnectPipe(g_hPipes.hPotsRx[slic], OTX_PIPE_OUTPUT); 
    DemoCheckRc(g_hPipes.hPotsRx[slic], result);
    result = OtxDrvDisconnectPipe(g_hPipes.hPotsTx[slic], OTX_PIPE_INPUT); 
    DemoCheckRc(g_hPipes.hPotsTx[slic], result);

    // Disconnect any burst connection that may have been setup
    setupCaptureLiData();
  }

  // These are no long needed
  g_liConDeviceNumber_1 = DEMO_NO_LI_CONNECTED;
  g_liConDeviceNumber_2 = DEMO_NO_LI_CONNECTED;


  DemoLogf("Done\n");

  return;
}


void DemoPrintEvent( OtxEventDataS *pEventData )
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_SLIC:
    // Processed in DemoSlicEventFunction()
    break;

    case OTX_DEVICE_BURST:
    {
      switch (pEventData->m_nCode) {
      case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
        DemoBurstReceiver(pEventData->m_nRequestId);
        break;
      case OTX_BURST_SENDER_EC_BUF_DEPLETED:
        DemoBurstSender(pEventData->m_nRequestId);
        break;
      default:
        otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        }
        break;
      }
      
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}


void DemoPrintTimeStamp( OTX_DATETIME nTimeStamp )
/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
  #endif

  // Convert to seconds since Jan 1, 1970 
  // (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds 
  // (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = (nTimeStamp.nUint64 / 10000) - 
                         (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}


void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{

  OTX_RESULT otxRc;
  char msgStr[256];
  OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
  // Size in bytes of attribute value 
  // (Get value with <f OtxDrvGetAttributeValue>)
  OTX_UINT32      cbAttribValue;  
  OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
  OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
  OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
    
  if ( OTX_FAILURE(result) ) {
    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } 
    else {      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, 
                                     OTX_ATTR_DEVICE_TYPE, 
                                     &eAttribType, 
                                     &cbAttribValue, 
                                     &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, 
                                     OTX_ATTR_DEVICE_NO, 
                                     &eAttribType, 
                                     &cbAttribValue, 
                                     &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", 
               OTX_RESULT_CODE(result), 
               OTX_RESULT_SOURCE_ID(result), 
               OTX_RESULT_LINE_NUM(result), 
               hDevice, 
               (int) eDeviceType, 
               (int)nDeviceNumber, 
               msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if(g_bDemoExitQuestion == OTX_FALSE) {
    return;
  }

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("\nDo you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
    DemoLogf("\n");
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }

  DemoLogf("\n");

}


/*******************************************************************************
  @func
    Function that keeps reseting the Watchdog timer on the board
*******************************************************************************/
void  *DemoWatchDogFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_BOOL bEnabled;

  g_bWatchDogThreadActive = OTX_TRUE;

  DemoLogf("DemoWatchDogFunction Thread started\n");

  while (g_bWatchDogThreadActive) {

    result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  &bEnabled); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (!bEnabled) {
      DemoLogf("*********** Watchdog Timer timed out. Application must be restarted *********\n");
      OsaThreadSleep(3000); // in milliseconds
      continue;
    }

    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_RESET,  OTX_TRUE /* Any value will work */); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    OsaThreadSleep(100); // in milliseconds
  }

  DemoLogf("DemoWatchDogFunction Thread terminating\n");
  return 0;

}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoSlicEventFunction( void *lpParameter  /* thread data */ ) 
{
  OTX_UINT32 i;
  OTX_UINT32 slicIndexNumber;
  OTX_UINT32 hookStatus;
  OTX_RESULT result;
  OtxEventDataS eventData;

  OsaThreadSleep(300); // Give hardware time to startup and initialize

  for(i = 0; i < DEMO_LI_MAX; i++) {
    OtxSlicGetHookState( g_hPhysDevs.hSlicLi[i], &hookStatus );
    DemoLogf("Slic[%d] Hook Status %d \n", i, hookStatus);
    
    if(hookStatus == OTX_SLIC_CS_ONHOOK) {
      setState(ONHOOK, i);  
    }
    else {
      setState(OFFHOOK, i);  
    }

  }

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); 
      DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event

        switch (eventData.m_eDeviceType) {
          case OTX_DEVICE_LI_SLIC:
            // determine slic index
            slicIndexNumber = eventData.m_nSourceId;
            // Determine the event
            switch (eventData.m_nCode) {
              // Pulse Metering Active Timer Interrupt
              case (OTX_SLIC_PULSE_INACTIVE):
              // Pulse Metering Inactive Timer Interrupt
              case (OTX_SLIC_PULSE_ACTIVE):
              // Ring Trip Interrupt
              case (OTX_SLIC_RING_TRIP):  
              // Loop Closure Transition Interrupt
              case (OTX_SLIC_LOOP_STATE):  
              // Indirect Register Access Serviced
              case (OTX_SLIC_INDIRECT_ACCESS_RDY):
              // Common Mode Calibration Error
              case (OTX_SLIC_ERROR_CM_CAL):  
              // Continue Ringing Active Timer
              case (OTX_SLIC_RING_INACTIVE_CONT):  
                //DemoLogf("OTX_SLIC_RING_INACTIVE_CONT\n");
                /* Valid but not used in this application
                DemoLogf("Slic[%d] Valid Event: %ld \n", 
                eventData.m_nSourceId, 
                eventData.m_nCode);
                */
                setState(NO_ACTION_NEEDED, slicIndexNumber);
                break;
                // Oscillator 1 Active Timer Interrupt
              case (OTX_SLIC_OSC1_INACTIVE):
                // Oscillator 1 Inactive Timer Interrupt
                DemoLogf("OTX_SLIC_OSC1_INACTIVE\n");
                setState(NO_ACTION_NEEDED, slicIndexNumber);
                break;
              case (OTX_SLIC_OSC1_ACTIVE):  
                // Oscillator 2 Active Timer Interrupt
                DemoLogf("OTX_SLIC_OSC1_ACTIVE\n");
                setState(NO_ACTION_NEEDED, slicIndexNumber);
                break;
              case (OTX_SLIC_OSC2_INACTIVE):
                // Oscillator 2 Inactive Timer Interrupt
                DemoLogf("OTX_SLIC_OSC2_INACTIVE\n");
                setState(NO_ACTION_NEEDED, slicIndexNumber);
                break;
              case (OTX_SLIC_OSC2_ACTIVE): 
                DemoLogf("OTX_SLIC_OSC2_ACTIVE\n");
                setState(NO_ACTION_NEEDED, slicIndexNumber);
                break;
                
              // Power Alarm Q1 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ1):  
              // Power Alarm Q2 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ2):  
              // Power Alarm Q3 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ3):  
              // Power Alarm Q4 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ4):  
              // Power Alarm Q5 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ5):  
              // Power Alarm Q6 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ6):

                // Slic reported error. Driver will try and correct
                // but if it continues device will be  disabled.
                setState(INITIALIZING, slicIndexNumber);
                DemoLogf("\n***  WARNING: Slic[%d] Power Alarm!\n\n", 
                         eventData.m_nSourceId);
                break;

              // DTMF Tone Detected
              case (OTX_SLIC_DTMF_DETECT):
                // DTMF digit is passed with event
                g_rawDigit = (OTX_UINT8) eventData.m_nParam;
                setState(DTMF_TRANISTION, slicIndexNumber);
                break;

              // Slic indicates OFFHOOK condition
              case (OTX_SLIC_OFF_HOOK):  
                setState(OFFHOOK, slicIndexNumber);
                break;
              // Slic indicates ONHOOK condition
              case (OTX_SLIC_ON_HOOK):  
                setState(ONHOOK, slicIndexNumber);
                break;

              // Ringing Active Timer Interrupt
              case (OTX_SLIC_RING_INACTIVE):
              // Ringing Inactive Timer Interrupt.
              case (OTX_SLIC_RING_ACTIVE):  
                setState(RINGING, slicIndexNumber);
                break;

            default:
              DemoLogf("Slic[%d] Unknown Event: %ld \n", 
                       eventData.m_nSourceId, 
                       eventData.m_nCode);

              break;
            }
            
            if(g_RunNormalMode == OTX_TRUE) {
              stateMachine(slicIndexNumber);
            }
            else {
              slicLoopTest(slicIndexNumber);
            }

            break;
          default:
            break;
        }

        // Printf the event data.
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}


/*******************************************************************************
  @func
    POTS Li state Machine. 
    This function will connect a single pair of FXS lines. It is used to 
    demonstrate the processing needed to configure the FXS interface.
*******************************************************************************/
void stateMachine(OTX_UINT32 deviceNumber)
{
  OTX_RESULT result;
  OTX_UINT32 slicState;
  OTX_UINT32 hookStatus;
  
  if (chipDataState[deviceNumber] != NO_ACTION_NEEDED) {
    DemoLogf("Li %d Current state %s [%d]\n", 
      deviceNumber,
      g_sqszSlicDeviceCallState[chipDataState[deviceNumber]], 
      chipDataState[deviceNumber]);
  }

  switch (chipDataState[deviceNumber])      
  {
           
  case CALL_BACK:
    break; 
  case CALLBACK_PROGRAMMED:
    break; 
  case MAKE_OFFHOOK: // dialtone 
    break; 
  case MAKE_BUSY:
    break;
  case MAKE_RING_BACK:
    break;
  case MAKE_REORDER:
    break;
  case MAKE_CONGESTION:
    break;
  case STOP_RINGING:
    break;
  case LOOP_TRANSITION:                    
    break;

  case NO_ACTION_NEEDED:
    // Need to set the current state back to previous state.
    chipDataState[deviceNumber] = chipDataPreviousState[deviceNumber];
    break;

  case STATE_CHANGE:
    OtxSlicGetHookState( g_hPhysDevs.hSlicLi[deviceNumber], &hookStatus );
    DemoLogf("Slic[%d] Hook Status %d \n", deviceNumber, hookStatus);
    
    if(hookStatus == OTX_SLIC_CS_ONHOOK) {
      setState(ONHOOK, deviceNumber);  
    }
    else {
      setState(OFFHOOK, deviceNumber);  
    }
    break; 

  case ONHOOK:
//    g_bDemoExitQuestion = OTX_TRUE; // enable demo exit question

    // Check to make sure the user has the phone on hook.
    OtxSlicGetHookState( g_hPhysDevs.hSlicLi[deviceNumber], &hookStatus );
    if(hookStatus == OTX_SLIC_CS_ONHOOK) {
      // The phone is still on hook it must be true
      if((g_liConDeviceNumber_1 != DEMO_NO_LI_CONNECTED) &&
         (g_liConDeviceNumber_2 != DEMO_NO_LI_CONNECTED)) {

        if((chipDataPreviousState[g_liConDeviceNumber_2] == RINGING) ||
           (chipDataPreviousState[g_liConDeviceNumber_2] == ONHOOK)) {
          // Did not connect the two phones
          DemoLogf("Li %d did not answer\n", g_liConDeviceNumber_2);

          // Stop ring on POTS Li
          result = OtxSlicStopRingState(g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2], 
                                        &slicState );  
          DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2], result);

          DemoLogf("  Li %d current driver state %s [%d]\n", 
                    g_liConDeviceNumber_2,
                    g_sqszSlicDeviceCallState[slicState], 
                    slicState);

          // Put this device into stop ring state
          setState(STOP_RINGING, g_liConDeviceNumber_2);

          // Line 2 is RINGING no answer
          g_liConDeviceNumber_1 = DEMO_NO_LI_CONNECTED;
          g_liConDeviceNumber_2 = DEMO_NO_LI_CONNECTED;
        }
        else {
          // 1st line onhook will break connection
          DemoDisconnectLiDevices();
        }
      }
      else {
        DemoLogf("Li %d ONHOOK\n", deviceNumber);
      }

      setState(ONHOOK, deviceNumber);

    }
    else {
      DemoLogf("Slic[%d] Flash Hook %d \n", deviceNumber, hookStatus);
      setState(FLASHHOOK, deviceNumber);  
    }

    break;

  case OFFHOOK: // change in slic state, now off hook
//    g_bDemoExitQuestion = OTX_FALSE; // disable demo exit question

    if((chipDataPreviousState[deviceNumber] == MAKE_OFFHOOK) ||
       (chipDataPreviousState[deviceNumber] == OFFHOOK)      ||
       (chipDataPreviousState[deviceNumber] == DTMF_TRANISTION)) {

      DemoLogf("Li %d is already Off Hook (FLASH HOOK?)\n", deviceNumber);

      setState(MAKE_OFFHOOK, deviceNumber);
    }
    else
    if(chipDataPreviousState[deviceNumber] == RINGING) {
      if((g_liConDeviceNumber_1 == DEMO_NO_LI_CONNECTED) ||
         (g_liConDeviceNumber_2 == DEMO_NO_LI_CONNECTED)) {

        DemoLogf("Warning: Li %d indicates RINGING and went OFFHOOK"
                 " but the connection globals are not setup\n", 
                 deviceNumber); 
      }
      else {
        // Connect slic audio for device starting the call
        result = OtxSlicEnableAudio( g_hPhysDevs.hSlicLi[g_liConDeviceNumber_1], 
                                     &slicState );  
        DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConDeviceNumber_1], result);

        if(slicState == OTX_SLIC_CS_ONHOOK) {
          DemoLogf("Warning: Current state for Li %d OTX_SLIC_CS_ONHOOK\n", 
                   g_liConDeviceNumber_1);
        }      
        // Connect slic audio for device being called
        if(g_liConDeviceNumber_2 != deviceNumber) {
          DemoLogf("Warning: Li(%d) is ring but Li(%d) should be\n", 
                   deviceNumber, g_liConDeviceNumber_2);
        }
        result = OtxSlicEnableAudio( g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2], 
                                     &slicState );  
        DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2], result);

        if(slicState == OTX_SLIC_CS_ONHOOK) {
          DemoLogf("Warning: Current state for Li %d OTX_SLIC_CS_ONHOOK\n", 
                   g_liConDeviceNumber_2);
        }

        // Connect the two lines
        DemoConnectLiToLiDevices();
      }
      setState(MAKE_OFFHOOK, deviceNumber);
    }
    else 
    if((chipDataPreviousState[deviceNumber] == ONHOOK)  ||
       (chipDataPreviousState[deviceNumber] == STOP_RINGING))  {
      // enable the user dialtone
      result = OtxSlicEnableDialTone( g_hPhysDevs.hSlicLi[deviceNumber] );
      DemoCheckRc(g_hPhysDevs.hSlicLi[deviceNumber], result);

      chipDataRingCount[deviceNumber]  = 0;
      chipDataDigitCount[deviceNumber] = 0;
      setState(MAKE_OFFHOOK, deviceNumber);

      DemoLogf("Li %d OFFHOOK\n", deviceNumber);

    }
    else {
      DemoLogf("Li %d no state change for OFFHOOK event (FLASH HOOK?)\n", 
               deviceNumber);
    }
    break;
  case DTMF_TRANISTION:
    dtmfAction(g_rawDigit, deviceNumber);

    setState(DIGIT_DECODING, deviceNumber);        
    break;
  case CALLER_ID:
    setState(RINGING, deviceNumber);
    break;
  case RINGING:
    if(chipDataPreviousState[deviceNumber] == RINGING) {

      if(chipDataRingCount[deviceNumber] > 4) {

        chipDataRingCount[deviceNumber] = 0;

        DemoLogf("  Li %d Ring Count timeout\n", 
                 deviceNumber);

        // Stop ring on POTS Li
        result = OtxSlicStopRingState( g_hPhysDevs.hSlicLi[deviceNumber], 
                                       &slicState );  
        DemoCheckRc(g_hPhysDevs.hSlicLi[deviceNumber], result);

        DemoLogf("  Li %d current driver state %s [%d]\n", 
                  deviceNumber,
                  g_sqszSlicDeviceCallState[slicState], 
                  slicState);

        setState(STOP_RINGING, deviceNumber);

      }
      else {
        chipDataRingCount[deviceNumber]++;
      }
    }
    break;
  }

}


/*******************************************************************************
  @func
    POTS Li state Machine for testing the SLIC device in a continous loop.
    The following code will connect FXS 0-1, 2-3 and 4-5 when either  of 
    the pair is taken off hook.
*******************************************************************************/
void slicLoopTest(OTX_UINT32 deviceNumber)
{

  DemoLogf("Li %d Current state %s [%d]\n", 
            deviceNumber,
            g_sqszSlicDeviceCallState[chipDataState[deviceNumber]], 
            chipDataState[deviceNumber]);

  switch (chipDataState[deviceNumber])      
  {
           
  case NO_ACTION_NEEDED:
    // Need to set the current state back to previous state.
    chipDataState[deviceNumber] = chipDataPreviousState[deviceNumber];
  case CALL_BACK:
  case CALLBACK_PROGRAMMED:
  case MAKE_OFFHOOK: // dialtone 
  case MAKE_BUSY:
  case MAKE_RING_BACK:
  case MAKE_REORDER:
  case MAKE_CONGESTION:
  case STOP_RINGING:
  case LOOP_TRANSITION:                    
  case DTMF_TRANISTION:
  case CALLER_ID:
  case RINGING:
  case STATE_CHANGE:

    break; 

  case ONHOOK:

    slicLoopTestOnHook(deviceNumber);

    break;

  case OFFHOOK: // change in slic state, now off hook
    
    slicLoopTestOffHook(deviceNumber);

    break;

  }

}


/*******************************************************************************
  @func
    POTS Li state Machine for testing the SLIC device in a continous loop.
    The following code will connect FXS 0 to 1, 2 to 3 or 4 to 5 when either
    of the pair is placed on hook.
*******************************************************************************/
void slicLoopTestOnHook(OTX_UINT32 deviceNumber)
{
  OTX_UINT32 deviceMate;
  OTX_UINT32 slicState;
  OTX_UINT32 hookStatus;
  OTX_RESULT result;

  // FXS 0-1, 2-3 and 4-5 are pairs. 
  if(deviceNumber%2) { 
    deviceMate = deviceNumber - 1;
  }
  else {
    deviceMate = deviceNumber + 1;
  }

  DemoLogf("slicLoopTestOnHook Li %d - Li %d\n", deviceNumber, deviceMate);

  // Check to make sure the user has the phone on hook.
  OtxSlicGetHookState( g_hPhysDevs.hSlicLi[deviceNumber], &hookStatus );
  if(hookStatus == OTX_SLIC_CS_ONHOOK) {
    // The phone is still on hook it must be true.

    // Now check to see if the other Li was active
    if(chipDataPreviousState[deviceMate] == RINGING) {
      // Did not connect the two phones
      DemoLogf("  Li %d OnHook and Ringing\n", deviceMate);

    }
    else
    if(chipDataPreviousState[deviceMate] == MAKE_OFFHOOK) {
      DemoLogf("  Li %d is still Off Hook, but not ringing\n", 
                  deviceMate);
    }

    // Stop ring on POTS Li (always in this test)
    result = OtxSlicStopRingState(g_hPhysDevs.hSlicLi[deviceMate], 
                                  &slicState );  
    DemoCheckRc(g_hPhysDevs.hSlicLi[deviceMate], result);

    DemoLogf("  Li %d current driver state %s [%d]\n", 
              deviceMate,
              g_sqszSlicDeviceCallState[slicState], 
              slicState);

    // Put this device into stop ring state
    setState(STOP_RINGING, deviceMate);

    setState(ONHOOK, deviceNumber);

  }
  else {
    DemoLogf("   Li %d is actually OFF Hook but event was ONHOOK\n", 
             deviceNumber);

    if(chipDataPreviousState[deviceNumber] != MAKE_OFFHOOK) {
      // software did not keep up with state change correctly
      DemoLogf("  Li %d state is now set to OFFHOOK\n", deviceNumber);

      setState(OFFHOOK, deviceNumber);
    }
  }
}



/*******************************************************************************
  @func
    POTS Li state Machine for testing the SLIC device in a continous loop.
    The following code will connect FXS 0 to 1, 2 to 3 or 4 to 5 when either
    of the pair is placed off hook.
*******************************************************************************/
void slicLoopTestOffHook(OTX_UINT32 deviceNumber)
{
  OTX_UINT32 deviceMate;
  OTX_UINT32 slicState;
  OTX_RESULT result;

  // FXS 0-1, 2-3 and 4-5 are pairs. 
  if(deviceNumber%2) { 
    deviceMate = deviceNumber - 1;
  }
  else {
    deviceMate = deviceNumber + 1;
  }

  DemoLogf("slicLoopTestOffHook Li %d - Li %d\n", deviceNumber, deviceMate);


  if((chipDataPreviousState[deviceNumber] == MAKE_OFFHOOK) ||
     (chipDataPreviousState[deviceNumber] == OFFHOOK)      ||
     (chipDataPreviousState[deviceNumber] == DTMF_TRANISTION)) {

    DemoLogf("  Li %d is already Off Hook (FLASH HOOK?)\n", deviceNumber);

    setState(MAKE_OFFHOOK, deviceNumber);
  }
  else
  if(chipDataPreviousState[deviceNumber] == RINGING) {
    // Connect slic audio for device starting the call
    result = OtxSlicEnableAudio( g_hPhysDevs.hSlicLi[deviceMate], 
                                 &slicState );  
    DemoCheckRc(g_hPhysDevs.hSlicLi[deviceMate], result);

    if(slicState == OTX_SLIC_CS_ONHOOK) {
      DemoLogf("  Warning: Current state for Li %d OTX_SLIC_CS_ONHOOK\n", 
               deviceMate);
    }      
    // Connect slic audio for device being called
    result = OtxSlicEnableAudio( g_hPhysDevs.hSlicLi[deviceNumber], 
                                 &slicState );  
    DemoCheckRc(g_hPhysDevs.hSlicLi[deviceNumber], result);

    if(slicState == OTX_SLIC_CS_ONHOOK) {
      DemoLogf("  Warning: Current state for Li %d OTX_SLIC_CS_ONHOOK\n", 
               deviceNumber);
    }

    setState(MAKE_OFFHOOK, deviceMate);
    setState(MAKE_OFFHOOK, deviceNumber);
  }
  else 
  if((chipDataPreviousState[deviceNumber] == ONHOOK)  ||
     (chipDataPreviousState[deviceNumber] == STOP_RINGING))  {
    // enable the user dialtone. The user will not hear this because
    // the ringback tone is setup in the next step. But it is part of
    // the normal setup so it is here.
    result = OtxSlicEnableDialTone( g_hPhysDevs.hSlicLi[deviceNumber] );
    DemoCheckRc(g_hPhysDevs.hSlicLi[deviceNumber], result);

    OsaThreadSleep(3000); // Give a little time to dialtone

    chipDataRingCount[deviceNumber]  = 0;
    chipDataDigitCount[deviceNumber] = 0;
    setState(MAKE_OFFHOOK, deviceNumber);

    // start ringing other line
    g_liConDeviceNumber_1 = DEMO_NO_LI_CONNECTED; 
    g_liConDeviceNumber_2 = deviceMate;   // line that will ring
    // To make sure to pass check performed by DemoCallLineMakeRing()
    g_dtmfDeviceNumber    = deviceNumber; // ringback tone
    DemoCallLineMakeRing();
    g_liConDeviceNumber_1 = DEMO_NO_LI_CONNECTED;
    g_liConDeviceNumber_2 = DEMO_NO_LI_CONNECTED;

    setState(RINGING, deviceNumber);
  
  }
  else {
    DemoLogf("  Li %d no state change for OFF_HOOK event (FLASH HOOK?)\n", 
             deviceNumber);
  }
}



/*******************************************************************************
@func
  POTS Li set state
*******************************************************************************/
void setState(OTX_UINT32 newState, OTX_UINT32 deviceNumber)
{
  // Make sure that we do not try to set it to state that is handled 
  // by the driver assert (newState >= OTX_SLIC_CS_END_OF_STATES);
  if (newState < OTX_SLIC_CS_END_OF_STATES) {
    DemoLogf("**** Setting %d state\n", newState);
  }

  if(deviceNumber >= DEMO_LI_MAX) {
    DemoLogf("Error: deviceNumber %d not valid.  %d state\n", 
              deviceNumber, newState);
  }
  else {
    chipDataPreviousState[deviceNumber] = chipDataState[deviceNumber];
    chipDataState[deviceNumber] = newState;
  }
}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) 
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_ARNI_6X6_PCI;
  g_nCurrentLi            = CURRENT_LI;

  for (i = 1; i < argc; i++) {   // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } 
    else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);        
        } else if (*(argv[i]+1) == 'N') {   
          // Slic number 1 used for demo POTS connection
          g_nCurrentLi = strtoul(argv[i]+2, NULL, 0);
          if(g_nCurrentLi >= DEMO_LI_MAX){
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if        (OsaToUpper(*(argv[i] + 2)) == 'A') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'C') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6_FXS_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'D') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X3_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'J') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI;
          } else {            
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /N<slicNo>   : Slic number 0-5\n");
  DemoLogf(" /T[A|C|D|J]  : Board Type used for the demo, where\n");
  DemoLogf("                    A  = Use Arni-6-PCI-Plus Board\n");
  DemoLogf("                    C  = Use Arni-6-FXS-PCI-Plus Board\n");
  DemoLogf("                    D  = Use Arni-6x3-PCI-Plus Board\n");
  DemoLogf("                    J  = Use Arni-6-PCI Board\n");
}


static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  // Print a timestamp for each line as well
  if (g_bLibConnected)  {
    OTX_DATETIME nTimeStampMs;
    OTX_DATETIME nTimeStampSec;
    struct Osa_tm *ptCurrentDayAndTime;
    Osa_time_t nCurrentTimeSecs;
    OTX_DATETIME nTimeStamp;
    
    OtxDrvGetCurrentTime(&nTimeStamp);
    
    // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
    #if defined(__linux__)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
    #elif defined(WIN32)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
    #endif
    
    // Convert to seconds since Jan 1, 1970 
    // (so we can use this time with OsaStdLocalTime())
    nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
    
    // Calculate number of milliseconds 
    // (since number of seconds is not a very accurate way to report timestamp)
    nTimeStampMs.nUint64 = (nTimeStamp.nUint64 / 10000) - 
      (nTimeStampSec.nUint64 * 1000);
    
    nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
    ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
    
    if (ptCurrentDayAndTime) {
      fprintf(g_pDemoLogFile, "%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
        ptCurrentDayAndTime->tm_mon + 1,
        ptCurrentDayAndTime->tm_mday,
        ptCurrentDayAndTime->tm_hour,
        ptCurrentDayAndTime->tm_min,
        ptCurrentDayAndTime->tm_sec,
        nTimeStampMs.sUint32.nLow
        );
    } else {
      // Just skip the printing of the timestamp in case of an erroneous timestamp

      //printf("Error: Invalid timestamp (before Jan 1, 1970)\n");
      //fprintf(g_pDemoLogFile, "Error: Invalid timestamp (before Jan 1, 1970)\n");
    }
    
  }
  

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile, formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}


/*******************************************************************************
  @func
    User has dialed a phone number. Put the LI into the RINGING state.
*******************************************************************************/
void DemoCallLineMakeRing(void) {
  OTX_UINT32 slicState;
  OTX_RESULT result;

  if(g_liConDeviceNumber_1 != DEMO_NO_LI_CONNECTED) {
    DemoLogf("Demo supports only single connection\n");
    return;
  }
  if(g_liConDeviceNumber_2 == g_dtmfDeviceNumber) {
    // We are calling our own number
    DemoLogf("You just called your own number.\n");

    g_liConDeviceNumber_1 = DEMO_NO_LI_CONNECTED;
    g_liConDeviceNumber_2 = DEMO_NO_LI_CONNECTED;
    return;
  }

  // Device index starting the call
  g_liConDeviceNumber_1 =  g_dtmfDeviceNumber; 
  
  if(chipDataState[g_liConDeviceNumber_2] != ONHOOK) {
    DemoLogf("The current state of Li(%d) is not ONHOOK\n", 
             g_liConDeviceNumber_2);
  }

  // check to see if the caller ID is enabled
  if(g_bSendCallerId == OTX_TRUE) {
    DemoLogf("Sending caller ID Li(%d)\n", g_liConDeviceNumber_2);

    result = OtxSlicSendCallerID( g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2],
                                  &slicState);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2], result);
  }

  DemoLogf("Ring Li(%d)\n", g_liConDeviceNumber_2);

  chipDataRingCount[g_liConDeviceNumber_1] = 0;
  chipDataRingCount[g_liConDeviceNumber_2] = 0;

  result = OtxSlicStartRingState( g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2], 
                                  &slicState );
  DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConDeviceNumber_2], result);

  // Start the ring back on the caller's line
  result = OtxSlicEnableRingBackTone(g_hPhysDevs.hSlicLi[g_dtmfDeviceNumber]);
  DemoCheckRc(g_hPhysDevs.hSlicLi[g_dtmfDeviceNumber], result);

  return;
}


/*******************************************************************************
  @func
    Function used in phone list jump table to call SLIC 0
*******************************************************************************/
void DemoCallLine_0(void) { 
  // Device index being called
  g_liConDeviceNumber_2 =  0; 

  DemoCallLineMakeRing();
}


/*******************************************************************************
  @func
    Function used in phone list jump table to call SLIC 1
*******************************************************************************/
void DemoCallLine_1(void) { 
  // Device index being called
  g_liConDeviceNumber_2 =  1; 

  DemoCallLineMakeRing();
}


/*******************************************************************************
  @func
    Function used in phone list jump table to call SLIC 2
*******************************************************************************/
void DemoCallLine_2(void) { 
  // Device index being called
  g_liConDeviceNumber_2 =  2; 

  DemoCallLineMakeRing();
}


/*******************************************************************************
  @func
    Function used in phone list jump table to call SLIC 3
*******************************************************************************/
void DemoCallLine_3(void) { 
  // Device index being called
  g_liConDeviceNumber_2 =  3; 

  DemoCallLineMakeRing();
}


/*******************************************************************************
  @func
    Function used in phone list jump table to call SLIC 4
*******************************************************************************/
void DemoCallLine_4(void) { 
  // Device index being called
  g_liConDeviceNumber_2 =  4; 

  DemoCallLineMakeRing();
}


/*******************************************************************************
  @func
    Function used in phone list jump table to call SLIC 5
*******************************************************************************/
void DemoCallLine_5(void) { 
  // Device index being called
  g_liConDeviceNumber_2 = 5; 

  DemoCallLineMakeRing();
}


/*******************************************************************************
  @func
    Function used in phone list jump table to send busy tone back to the SLIC
*******************************************************************************/
void DemoBusy(void) { 
  OTX_RESULT result;
  g_bDemoExitQuestion = OTX_FALSE; // disable demo exit question

  result = OtxSlicEnableBusyTone(g_hPhysDevs.hSlicLi[g_dtmfDeviceNumber]);
  DemoCheckRc(g_hPhysDevs.hSlicLi[g_dtmfDeviceNumber], result);
}


/*******************************************************************************
  @func
    Function used in phone list jump table to send reorder tone back to the SLIC
*******************************************************************************/
void DemoReOrder(void) { 
  OTX_RESULT result;
  g_bDemoExitQuestion = OTX_TRUE; // enable demo exit question

  result = OtxSlicEnableReorderTone(g_hPhysDevs.hSlicLi[g_dtmfDeviceNumber]);
  DemoCheckRc(g_hPhysDevs.hSlicLi[g_dtmfDeviceNumber], result);
}


/*******************************************************************************
  @func
    Function used in phone list jump table for a non-listed number
*******************************************************************************/
void DemoNotListed(void) { 

  ;
}



/*******************************************************************************
    Phone list jump table setup
*******************************************************************************/
typedef struct  {
    char * phoneNumber ;
    char * functionName;
    void (*action)();
}tNUMBER;

tNUMBER phoneNumbers[] = {
    
    { "3210","Busy Tone",    DemoBusy },
    { "101" ,"Busy Tone 101",DemoBusy },
    { "9876","Reorder Tone", DemoReOrder },
    { "1000","Call Line 0",  DemoCallLine_0 },
    { "1001","Call Line 1",  DemoCallLine_1 },
    { "1002","Call Line 2",  DemoCallLine_2 },
    { "1003","Call Line 3",  DemoCallLine_3 },
    { "1004","Call Line 4",  DemoCallLine_4 },
    { "1005","Call Line 5",  DemoCallLine_5 },
    { "9726640100","Call Line 4",  DemoCallLine_4 },
    { "", "Not Displayed!",  DemoNotListed }   // Marks end of list 

};


/*******************************************************************************
  @func
    Function used to display the phone numbers 
*******************************************************************************/
void printPhoneNumbers(void)
{
  OTX_UINT32 i = 0;
  DemoLogf("\n");
  DemoLogf("\n\t \t \t P H O N E  D I R E C T O R Y\n ");
  while(*phoneNumbers[i].phoneNumber)
  {
    DemoLogf("\n \t %s \t %s",
            phoneNumbers[i].phoneNumber,
            phoneNumbers[i].functionName);
    i++;
  }
  DemoLogf("\n");
  DemoLogf("\n\t \t \t P H O N E  D I R E C T O R Y\n ");
  DemoLogf("\n");
}


/*******************************************************************************
  @func
    Function used to search the phone list for a valid number 
*******************************************************************************/
void (*findNumber(OTX_UINT32 deviceNumber))()
{
  static  int currentNumber;

  OTX_UINT8 tableR; // Row
  OTX_UINT32 tableC; // Column

  /*
  To find the number a while loop is start at the last place the program  
  left off in the phone number table and continues until either there is  
  a match with fewer digits than a table entry, or there is a match with   
  an entire entry, or there is no match and the end of the table is reached.
  */

  tableR = currentNumber;
  tableC = chipDataDigitCount[deviceNumber] - 1;

  while(phoneNumbers[tableR].phoneNumber[tableC] !=0)
  {
    if( phoneNumbers[tableR].phoneNumber[tableC] == 
        chipDataDtmfDigits[deviceNumber][tableC] )
    {
      tableC++;

      if(tableC == chipDataDigitCount[deviceNumber])
      {
        if(phoneNumbers[tableR].phoneNumber[tableC] == 0)
        {
          currentNumber = 0;
          tableC = 0;
          chipDataDigitCount[deviceNumber] = 0;

          return  phoneNumbers[tableR].action;
        }                            
        else
        {    
          return NULL ;
        }
      }
    } 
    else
    {
      tableR = ++currentNumber;
      tableC = 0;
    }
  }

  currentNumber = 0;
  chipDataDigitCount[deviceNumber] = 0;

  return &DemoReOrder ;
}

/*******************************************************************************

  @func
  Demo Collect the DTMF digits

  @comm
  The applications needs to keep a record of the digits received the driver
  does not perform any data buffering.

*******************************************************************************/
void dtmfAction(OTX_UINT8 rawDigit, OTX_UINT32 deviceNumber)
{  
  void (*funct)();
  char ascii;

  setState(DIGIT_DECODING, deviceNumber);

  g_dtmfDeviceNumber = deviceNumber;

  //asciiChar= '0' + digit;
  switch(rawDigit)
  {
  case 0xA :
    ascii = '0';
    break;
  case 0xB:
    ascii = '*';
    break;
  case 0xC:
    ascii = '#';
    break;
  default:
    ascii= '0' + rawDigit;
    break;
  }

  if(chipDataDigitCount[deviceNumber] < NUMBER_OF_DTMF_DIGITS)
  {
    chipDataDtmfDigits[deviceNumber][chipDataDigitCount[deviceNumber]] = ascii; 
    chipDataDigitCount[deviceNumber]++;
    chipDataDtmfDigits[deviceNumber][chipDataDigitCount[deviceNumber]]= 0;

    DemoLogf("Value= 0x%01x  String collected \"%s\" \n", 
             rawDigit, 
             &chipDataDtmfDigits[deviceNumber]);

    funct = findNumber(deviceNumber);
    if(funct) 
    {
      funct();
    }
  }
}


/*******************************************************************************

  @func
  Demo enable the caller ID and ringer on all six SLIC devices

*******************************************************************************/
void ringAllLines(void)
{
  OTX_INT32 slic;
  OTX_UINT32 slicState;
  OTX_RESULT result;

  for(slic = DEMO_LI_MAX - 1; slic >= 0; slic--) {
    // check to see if the caller ID is enabled
    if(g_bSendCallerId == OTX_TRUE) {
      if(chipDataState[slic] != ONHOOK) {
        DemoLogf("The current state of Li(%d) is not ONHOOK\n", 
                 slic);
      }

      DemoLogf("Enable caller ID Li(%d)\n", slic);

      OsaThreadSleep((OTX_UINT32)3000L); 

      result = OtxSlicSendCallerID( g_hPhysDevs.hSlicLi[slic],
                                    &slicState);
      DemoCheckRc(g_hPhysDevs.hSlicLi[slic], result);

      if(slicState != OTX_SLIC_CS_ONHOOK) {
        DemoLogf("The current state of Li(%d) is not OTX_SLIC_CS_ONHOOK\n", 
                 slic);
      }
    }
    DemoLogf("Enable ringing Li(%d)\n", slic);

    result = OtxSlicStartRingState( g_hPhysDevs.hSlicLi[slic], 
                                    &slicState );
    DemoCheckRc(g_hPhysDevs.hSlicLi[slic], result); 
    
    OsaThreadSleep((OTX_UINT32)1000L); 
  }


  DemoLogf("Ring All Liines\n");

  return;
}



/*******************************************************************************

  @func
  Demo configures the burst device to capture data on the calling FXS line.

*******************************************************************************/
void setupCaptureLiData(void)
{
  OTX_RESULT result;
  OTX_INT32 nHwy;
  OTX_INT32 nTs;
  OTX_UINT8 sqTimeSlotMask[32];

  static OTX_HANDLE hPipeRecSlic = 0;

  // Check to see if we have setup a previous connection
  if(g_bBurstEnabled == OTX_TRUE) {
    DemoLogf("Cleanup previous burst connection\n");

    // Close burst (RECEIVE) pipe connection
    result = OtxDrvDisconnectPipe(hPipeRecSlic, OTX_PIPE_OUTPUT); 
    DemoCheckRc(hPipeRecSlic, result);
    result = OtxDrvDisconnectPipe(hPipeRecSlic, OTX_PIPE_INPUT); 
    DemoCheckRc(hPipeRecSlic, result);

    // Need to cleanup this connection 
    result = OtxDrvClose(hPipeRecSlic); 
    DemoCheckRc(hPipeRecSlic, result);
    hPipeRecSlic = 0;

    // Close burst (SEND) pipe connection
    if(g_hPipeSendSlic != OTX_INVALID_HANDLE_VALUE) {
      result = OtxDrvDisconnectPipe(g_hPipeSendSlic, OTX_PIPE_OUTPUT); 
      DemoCheckRc(g_hPipeSendSlic, result);
      result = OtxDrvDisconnectPipe(g_hPipeSendSlic, OTX_PIPE_INPUT); 
      DemoCheckRc(g_hPipeSendSlic, result);

      // Need to cleanup this connection 
      result = OtxDrvClose(g_hPipeSendSlic); 
      DemoCheckRc(g_hPipeSendSlic, result);
      g_hPipeSendSlic = OTX_INVALID_HANDLE_VALUE;
    }

    // Handle to the physical Burst device.
    result = OtxDrvClose(g_hPhysDevs.hBurst); 
    DemoCheckRc(g_hPhysDevs.hBurst, result);  
    g_hPhysDevs.hBurst = OTX_INVALID_HANDLE_VALUE;

    g_bBurstEnabled = OTX_FALSE;
  }

  if( (DEMO_NO_LI_CONNECTED == g_liConDeviceNumber_1) || 
      (DEMO_NO_LI_CONNECTED == g_liConDeviceNumber_2) )  {
    DemoLogf("\n   FXS device not currently connected\n\n");
    return;
  }

  g_bBurstEnabled = OTX_TRUE;

  // Open Burst device
  // Get a handle to the physical Burst device.
  DemoLogf("Open OTX_DEVICE_BURST\n");
  if(g_hPhysDevs.hBurst == OTX_INVALID_HANDLE_VALUE) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, 
                                      OTX_DEVICE_BURST, 
                                      0, 
                                      0, 
                                      g_hEventQueue, 
                                      &g_hPhysDevs.hBurst); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  }

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, 
                                         OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  
                                         DEMO_BURST_BUF_SIZE); 
  DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, 
                                         OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    
                                         DEMO_BURST_BUF_SIZE); 
  DemoCheckRc(g_hPhysDevs.hBurst, result);


  // Also connect the FXS SLIC samples to the burst via a TSS connection
  result = OtxDrvCreatePipe(64, &hPipeRecSlic);
  DemoCheckRc(hPipeRecSlic, result);

  memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
  // Each slic will have a fixed location within the data stream.
  // Slic0 will be at TS0, Slic1 will be at TS3... Slic5 will be at TS15
  DemoLogf("  Connecting FXS %d to burst device\n", g_liConDeviceNumber_1);
  
  nTs = g_liConDeviceNumber_1 * 3; // calculate time slot
  nHwy = 6;    // SLIC Highway
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipeRecSlic, 
                             OTX_PIPE_INPUT, 
                             g_hPhysDevs.hTss,  
                             nHwy, 
                             sqTimeSlotMask, 
                             32); 
  DemoCheckRc(hPipeRecSlic, result);




  memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
  nTs = 0;
  nHwy = 8;    // BURST highway
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipeRecSlic, 
                             OTX_PIPE_OUTPUT, 
                             g_hPhysDevs.hTss,  
                             nHwy, 
                             sqTimeSlotMask, 
                             32); 
  DemoCheckRc(hPipeRecSlic, result);
}


/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;
  OTX_INT32 nTsToSendDirect = 0;
  OTX_INT32 nTsToSendDelayed = 1;
  int cbSingleTsBuf;
  int byteCount = 0;
  int x;

  // validate parameters
  if (nSourceId != 0) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstSender.pTxFile) {
    g_sBurstDevs.mBurstSender.pTxFile = fopen(DEMO_FXS_INPUT_FILE, "rb");
    if (g_sBurstDevs.mBurstSender.pTxFile == NULL) {
      DemoLogf("DemoBurstSender: Could not open file %s. "
               "Disabling the burst device\n", 
               DEMO_FXS_INPUT_FILE);
      OtxDrvDisable(g_hPhysDevs.hBurst, 5000, 2000);
      return;
    }
    
    g_sBurstDevs.mBurstSender.bFileLoop = OTX_TRUE;

    g_sBurstDevs.mBurstSender.bActive = OTX_TRUE;

    DemoLogf("Burst Sender started\n");

    memset(&g_sBurstDevs.mBurstSender.sqBufCurr[0], 
           DEMO_U_LAW_SILENCE, 
           sizeof(g_sBurstDevs.mBurstSender.sqBufCurr) / sizeof(g_sBurstDevs.mBurstSender.sqBufCurr[0]));
    memset(&g_sBurstDevs.mBurstSender.sqBufPrev[0], 
           DEMO_U_LAW_SILENCE, 
           sizeof(g_sBurstDevs.mBurstSender.sqBufPrev) / sizeof(g_sBurstDevs.mBurstSender.sqBufPrev[0]));

  }
  

  if (!g_sBurstDevs.mBurstSender.bActive) {
    // we're not active, so just return
    return;
  }

  if (!g_sBurstDevs.mBurstSender.pTxFile) {
    // No file to send, so just return
    return;
  }


  // Prime the transmit buffer with DEMO_U_LAW_SILENCE (0xFF)
  memset(&g_sBurstDevs.mBurstSender.sqTxBuf[0], 
         DEMO_U_LAW_SILENCE, 
         sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]));

  // Copy the previous buffer so we can use its samples for the delayed sample buffer
  memcpy(&g_sBurstDevs.mBurstSender.sqBufPrev[0], 
         &g_sBurstDevs.mBurstSender.sqBufCurr[0], 
         sizeof(g_sBurstDevs.mBurstSender.sqBufPrev) / sizeof(g_sBurstDevs.mBurstSender.sqBufPrev[0]));
  
  cbSingleTsBuf = fread(
    &g_sBurstDevs.mBurstSender.sqBufCurr,
    sizeof(g_sBurstDevs.mBurstSender.sqBufCurr[0]),
    sizeof(g_sBurstDevs.mBurstSender.sqBufCurr) / sizeof(g_sBurstDevs.mBurstSender.sqBufCurr[0]),
    g_sBurstDevs.mBurstSender.pTxFile
    );
  
  
  // Dispurse the single TS to be transmitted in the whole transmit buffer 
  // for the SendDirect timeslot on the burst highway
  byteCount = 0;
//  allTs     = g_sBurstDevs.mBurstSender.nTsToSend;
  for ( x = nTsToSendDirect; (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); x += 32) {
//    for( allTs = 0; allTs < 32; allTs++) {
//      g_sBurstDevs.mBurstSender.sqTxBuf[x + allTs] = g_sBurstDevs.mBurstSender.sqBufCurr[byteCount];
//    }
    g_sBurstDevs.mBurstSender.sqTxBuf[x] = g_sBurstDevs.mBurstSender.sqBufCurr[byteCount];
    byteCount++;
  }

  if(byteCount != cbSingleTsBuf) {
    DemoLogf("DemoBurstSender[%d]: Error in byteCount\n");
  }
  g_sBurstDevs.mBurstSender.cbTxBuf = cbSingleTsBuf * 32;

  byteCount = 0;
  for ( x = nTsToSendDelayed; (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); x += 32) {
      g_sBurstDevs.mBurstSender.sqTxBuf[x] = g_sBurstDevs.mBurstSender.sqBufCurr[byteCount++];
  }
  
  if (!feof(g_sBurstDevs.mBurstSender.pTxFile)) {
    assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);
  } else {
    DemoLogf("DemoBurstSender[%d]: Reached EOF\n", nSourceId);
  }
    
  if (g_sBurstDevs.mBurstSender.cbTxBuf > 0) {    
    // echo if we're in verbose mode
    if (g_sBurstDevs.mBurstSender.bVerbose) {
      int x;
      
      DemoLogf("DemoBurstSender[%d]: Transmitting %d bytes\n", nSourceId, g_sBurstDevs.mBurstSender.cbTxBuf);
      
      for (x = 0; (x < 32) && (x < (OTX_INT32) sizeof(g_sBurstDevs.mBurstSender.sqTxBuf)); x++) {
        // display the highway's currentLi's data
        if ((x % 8) == 0) DemoLogf("\n");
        DemoLogf("%.2X ", g_sBurstDevs.mBurstSender.sqTxBuf[x]);
      }
      DemoLogf("\n");
      DemoLogf("\n");
    }
    
    ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_sBurstDevs.mBurstSender.sqTxBuf, &ui32);
    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);
    
    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).

    if (!(ui32 == g_sBurstDevs.mBurstSender.cbTxBuf)) {
      DemoLogf("ui32(%d) != g_sBurstDevs.mBurstSender.cbTxBuf(%d)\n", ui32, g_sBurstDevs.mBurstSender.cbTxBuf);
    }
    
  } else {
    
    if (g_sBurstDevs.mBurstSender.bFileLoop) {
      if (g_sBurstDevs.mBurstSender.bVerbose) {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
      }
      rewind(g_sBurstDevs.mBurstSender.pTxFile);
    } else {
      DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
      fclose(g_sBurstDevs.mBurstSender.pTxFile);
      g_sBurstDevs.mBurstSender.pTxFile = NULL;

      g_sBurstDevs.mBurstSender.bActive = OTX_FALSE;
    }
  }
  
  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstReceiver.bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, 
                             nSourceId, 
                             buf, 
                             &ui32, 
                             &nFrameRefId, 
                             &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  // (since we set the trigger amount to the buffer size, there should
  // have been room to buffer this, so these two should be the same).
  if(!(ui32 == sizeof(buf))) {
    DemoLogf("\n");
    DemoLogf("Error: BurstReceiver[%d] Buffer Size(len=%d) Should be %d\n", 
             nSourceId, 
             sizeof(buf), 
             ui32);
    DemoLogf("\n");
  }

  // echo if we're in verbose mode
  if (g_sBurstDevs.mBurstReceiver.bVerbose) {
    int x;

    DemoPrintTimeStamp(dtFrameRefTime);

    // note: we only display the least significant 
    //       32 bits of the frame reference id.
    DemoLogf("BurstReceiver[%d] (len=%.5d) ID:%08x ", 
             nSourceId, 
             sizeof(buf), 
             (OTX_UINT32) (nFrameRefId & 0xffffffff));

    for (x = 0; (x < 32) && (x < (OTX_INT32) sizeof(buf)); x++) {
      // display the highway's currentLi's data
      if ((x % 8) == 0) DemoLogf("\n");
      DemoLogf("%.2X ", buf[x]);
    }
    DemoLogf("\n");
    DemoLogf("\n");
  }

  // write data to a file if desired
  if (g_sBurstDevs.mBurstReceiver.pRxFile) {
    int cbWritten;

    if (g_sBurstDevs.mBurstReceiver.nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, 
                         sizeof(buf[0]), 
                         sizeof(buf), 
                         g_sBurstDevs.mBurstReceiver.pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file."
                 "  (nFrameLen=%d, cbWritten=%d) \n", 
                 nSourceId, 
                 sizeof(buf), 
                 cbWritten);
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      int x;

      byteCount = 0;
      x = g_sBurstDevs.mBurstReceiver.nTsToRecord;
      for ( ; x < (OTX_INT32) sizeof(buf); x += 32) {
        singleTsBuf[byteCount++] = buf[x];
      }
      cbWritten = fwrite(singleTsBuf, 
                         sizeof(singleTsBuf[0]), 
                         byteCount, 
                         g_sBurstDevs.mBurstReceiver.pRxFile);
      if (cbWritten != byteCount) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file."
                 "  (byteCount=%d, cbWritten=%d) \n", 
                 nSourceId, 
                 byteCount, 
                 cbWritten);
      }
    }
  }
}


/*******************************************************************************

  @func

    Returns the state of the Burst device (enabled or disabled).

*******************************************************************************/
OTX_BOOL DemoIsBurstDeviceEnabled(void)
{
  OTX_RESULT result;
  OtxDeviceStateE  eDevState;

  if(g_bBurstEnabled == OTX_FALSE) {
    // Device has not been setup
    return OTX_FALSE; // The burst device is not enabled
  }

  result = OtxDrvGetState(g_hPhysDevs.hBurst, &eDevState); 
  DemoCheckRc(g_hPhysDevs.hBurst, result);

  if (eDevState == OTX_DEVICE_WORKING) {
    return OTX_TRUE; // The burst device is enabled
  } else {
    return OTX_FALSE; // The burst device is not enabled
  }
}


/*******************************************************************************
  @func

  Demo entry point for the application

*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_RESULT result;  // OTX Standard Return Code
  OTX_INT32 i;

  // Initialize the Demo call state variables
  for(i = 0; i < DEMO_LI_MAX ; i++) {
    chipDataRingCount[i] = 0;
    chipDataDigitCount[i] = 0;
    // Will verify this later
    chipDataState[i] = ONHOOK;
    chipDataNewState[i] = ONHOOK;
    chipDataPreviousState[i] = ONHOOK;

    g_hPipes.hPotsRx[i] = 0;  
  }
  g_liConDeviceNumber_1 = DEMO_NO_LI_CONNECTED;
  g_liConDeviceNumber_2 = DEMO_NO_LI_CONNECTED;

  g_hPhysDevs.hBurst = OTX_INVALID_HANDLE_VALUE;

  OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    exit(-1);
  }

  DemoLogf("%s, %s, Odin TeleSystems Inc.\n", argv[0], __DATE__);
  DemoLogf("A66Slic2, SLIC Demo for Arni-6x6-PCI and Arni-6x6-PCI-Plus\n");

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("Slic number  %d\n", g_nCurrentLi);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_REV_STR, 
                                       szVersionInfo, 
                                       255); 
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, 
                                         OTX_ATTR_LIB_TYPE_STATIC, 
                                         &bIsStaticLinkedLibrary); 
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_FILE_NAME, 
                                       szLibFileName, 
                                       64); 
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", 
             szVersionInfo, 
               szLibFileName, 
               bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library."
             "\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and "
             "OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  g_bLibConnected = OTX_TRUE;

  // Create Event Queues for Driver --> Application Communications
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Arni-16-PCI board
  DemoOpenPhysicalDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    // Size in bytes of attribute value 
    // (Get value with <f OtxDrvGetAttributeValue>)
    OTX_UINT32      cbAttribValue;  
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, 
                                    OTX_ATTR_BRD_CPLD_BUILD, 
                                    &eAttribType, 
                                    &cbAttribValue, 
                                    &bReadOnly); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, 
                                       OTX_ATTR_BRD_CPLD_BUILD, 
                                       pszCpldBuildNumbers, 
                                       cbAttribValue); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);

    free(pszCpldBuildNumbers);
  }

  // Enable Physical devices. Device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoSlicEventFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

#if 0
  // Enable the Watchdog Timer itself
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Start the Watchdog timer Reset Thread, which will keep the watchdog alive
  OsaThreadStart(DemoWatchDogFunction);
#else
  DemoLogf("Skipping enabling of Watchdog timer for now. Watchdog timer not tested with the Slics enabled.\n");
#endif
  

  DemoLogf("Press 'h' for help\n");

  // Execute user commands from console window
  {
    char commandChar;
    OTX_INT32 exitFlag = 0;

    while (!exitFlag) 
    {
      commandChar = OsaConsoleGetCh();
      exitFlag = DemoProcessUserCommand(commandChar);
    }

  }

  // Delete allocated resources
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

  This code is case sensitive.

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  char strBuf[256];
  OTX_UINT32 regReadAddL = 0;
  OTX_UINT32 regValueL = 0;
  OTX_INT32 i;
  OTX_RESULT result;
  unsigned char regValue = 0;


  DemoLogf("User command > %c (0x%x)\n", commandChar, commandChar);

  switch (commandChar)   {
  case '?':
  case 'h':
  case 'H':
    DemoLogf("\nCmd  Synopsis for PROSLIC Si321x\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit this menu\n");
    DemoLogf(" 0-5 Select active SLIC. #%d is now active\n", g_nCurrentLi);
    DemoLogf(" a   Set default TS within each Slic\n");
    DemoLogf(" b   Send polarity reversal\n");
    DemoLogf(" c   Set PCM Format\n");
    DemoLogf(" e   Configure Slic Audio Receive Path\n");
    DemoLogf(" i   Toggle ring cadence Standard/Non-standard\n");
    if(DemoIsBurstDeviceEnabled()) {
      DemoLogf(" M   Enable/Disable Burst (is Enabled)\n");
    }
    else {
      DemoLogf(" M   Enable/Disable Burst (is Disabled)\n");
    }
    DemoLogf(" N   Start Burst Receiver. Prompt for file to recv\n");
    DemoLogf(" O   Abort Burst Receivers\n");
    DemoLogf(" P   Burst send %s file to FXS 2\n", DEMO_FXS_INPUT_FILE);
    DemoLogf(" p   Print phone number list\n");
    DemoLogf(" r   Read Slic Direct Register\n");
    DemoLogf(" w   Write Slic Direct Register\n");
    DemoLogf(" R   Read FPGA Register\n");
    DemoLogf(" W   Write FPGA Register\n");
    DemoLogf(" s   Enable CID and ring LI\n");
    DemoLogf(" t   Set each SLIC to a different CID\n");
    DemoLogf(" u   Enter CID strings\n");
    DemoLogf(" U   Ring all %d SLIC Devices with CID\n", DEMO_LI_MAX);
    DemoLogf(" v   Enable Caller ID Send\n");
    DemoLogf(" V   Toggle verbose message print\n");
    DemoLogf(" !   Terminate Watchdog Timer Thread (Crash test)\n");
    DemoLogf(" f   Toggle Fail-over Relay for current Slic\n");
    DemoLogf("\n");
  break;

  case 'a':

    DemoLogf("Setting All 6 Slic's To Default Time Slots \n");

    for(i=0; i < 6; i++)
    {
      OTX_UINT16    pcmCount;
      unsigned char pcmCountLow;
      unsigned char pcmCountHigh;

      // Setup Slic Time Slots
      // Each slic will have a fixed location within the data stream.
      // Slic0 will be at TS0, Slic1 will be at TS3... Slic5 will be at TS15
      pcmCount = (OTX_UINT16)((i * 3) * 8);
      pcmCountLow  = pcmCount & 0xFF;
      pcmCountHigh = (pcmCount >> 8) & 0xFF;

      result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[i], 
                               2,      //TX_START_LO
                               pcmCountLow); 
      DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);

      result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[i], 
                               3,      //TX_START_HI
                               pcmCountHigh); 
      DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);

      result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[i], 
                               4,      //RX_START_LO
                               pcmCountLow); 
      DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);
  
      result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[i], 
                               5,      //RX_START_HI
                               pcmCountHigh); 
      DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);    
    }

    DemoLogf("Done\n");

  break;

  case 'b':
    {
    OTX_RESULT result;
    OTX_UINT32 slic;
    OTX_UINT32 slicState;

    OTX_INT32 delayCnt;

    DemoLogf("Enter number of the SLIC to toggle polarity: ");
    OsaConsoleGetStr(strBuf);

    slic = strtol(strBuf, NULL, 0);

    DemoLogf("Enter number of millisecont tick for reversal pulse : ");
    OsaConsoleGetStr(strBuf);

    delayCnt = strtol(strBuf, NULL, 0);

    result = OtxSlicSignalDisconnectState(g_hPhysDevs.hSlicLi[slic],
                                          delayCnt,
                                          &slicState);
    DemoCheckRc(g_hPhysDevs.hSlicLi[slic], result);


    DemoLogf("  Li %d current driver state %s [%d]\n", 
              slic,
              g_sqszSlicDeviceCallState[slicState], 
              slicState);

    DemoLogf("Polarity reversal complete SLIC %d\n", slic);

    }
  break;

  case 'c': // set compression
    {
      OTX_UINT32 compression;
      OTX_UINT32 i;

      for(i = 0; i < DEMO_LI_MAX; i++) {
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hSlicLi[i], 
                                               OTX_ATTR_SLIC_POTS_COMPRESSION_SET, 
                                               &compression); 
      
        if(compression == 0) {
          DemoLogf("SLIC(%d) is set to A-Law\n", i);
        } 
        else 
        if(compression == 1) {
          DemoLogf("SLIC(%d) is set to u-Law\n", i);
        }
        else 
        if(compression == 2) {
          DemoLogf("SLIC(%d) is set to Linear\n", i);
        }
        else {
          DemoLogf("SLIC(%d) Returned an unknown compression type\n", i);
        }
      
      }


      DemoLogf("\n");
      DemoLogf("   0 for A-Law\n");
      DemoLogf("   1 for u-Law\n");
      DemoLogf("   2 for Linear\n");
      DemoLogf("Enter : ");
      OsaConsoleGetStr(strBuf);


      if(strBuf[0] == '0') {
        compression = OTX_SLIC_CMP_SET_U_LAW;
      } 
      else 
      if(strBuf[0] == '1') {
        compression = OTX_SLIC_CMP_SET_A_LAW;
      }
      else 
      if(strBuf[0] == '2') {
        compression = OTX_SLIC_CMP_SET_LINEAR;
      }
      else {
        DemoLogf("\nNot a valid option\n\n");
        return 0;
      }

      for(i = 0; i < DEMO_LI_MAX; i++) {
        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hSlicLi[i], 
                                               OTX_ATTR_SLIC_POTS_COMPRESSION_SET, 
                                               compression); 
      }

      DemoLogf("Done\n");

    }
  break;

  case 'D':
    // This is only valid if the driver debug OtxDrv.sys is loaded
    DemoLogf("\nSLIC Force Power Error Recovery\n");
    result = OtxDrvDbgDump(g_hPhysDevs.hSlicLi[g_nCurrentLi], 500);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

  break;

  case 'e':
    DemoLogf("\nSLIC Receive Path Audio Configuration\n");

    {
      OTX_UINT32 exitMenu = 0;

      while( exitMenu == 0 ) {
        // Display help menu for user first
        exitMenu = DemoSlicReceiveConfigMenu('?');
        // Get user input
        OsaConsoleGetStr(strBuf);

        exitMenu = DemoSlicReceiveConfigMenu(strBuf[0]);
      }

    }

    DemoLogf("\nAudio Configuration done\n");

  break;

  case 'g': // Enable Ground Start signal on RING

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            64,
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("Reading SLIC(%d) register (64) = %X\n", g_nCurrentLi, regValueL);

    regValueL = regValueL & 0xF4;
    regValueL = regValueL | 0x03;

    DemoLogf("Enabled GS (64) = %X\n", regValueL);

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi], 
                             regReadAddL, 
                             regValue); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

  break;

  case 'G': // Remove Ground Start signal on RING

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            64,
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("Reading SLIC(%d) register (64) = %X\n", g_nCurrentLi, regValueL);

    regValueL = regValueL & 0xF4;
    regValueL = regValueL | 0x01;

    DemoLogf("Enabled GS (64) = %X\n", regValueL);

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi], 
                             regReadAddL, 
                             regValue); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

  break;

  case 'i': // toggle ring cadence
    {
      OTX_UINT32 nRingType;

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hSlicLi[g_nCurrentLi], 
                                             OTX_ATTR_SLIC_RING_TYPE_CONTROL, 
                                             &nRingType); 

      DemoLogf("SLIC(%d) ring cadence was %d\n", g_nCurrentLi, nRingType);

      // For now there is only two types of ring cadences
      if(nRingType == OTX_SLIC_CFG_SET_RING_0) {
        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hSlicLi[g_nCurrentLi], 
                                               OTX_ATTR_SLIC_RING_TYPE_CONTROL, 
                                               OTX_SLIC_CFG_SET_RING_1); 
        DemoLogf("Ring type is OTX_SLIC_CFG_SET_RING_1 (non-standard)\n");
      } else {
        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hSlicLi[g_nCurrentLi], 
                                               OTX_ATTR_SLIC_RING_TYPE_CONTROL, 
                                               OTX_SLIC_CFG_SET_RING_0); 
        DemoLogf("Ring type is OTX_SLIC_CFG_SET_RING_0 (standard)\n");
      }
    }
  break;

  // BURST COMMANDS (M, N, O)
  case 'M':
    if(DemoIsBurstDeviceEnabled()) {
      // By disabling the Burst device, we stop the process of transferring 
      // data from the hardware to the driver via DMA.
      DemoLogf("Disable g_hPhysDevs.hBurst\n");
      result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
      DemoCheckRc(g_hPhysDevs.hBurst, result);
    } 
    else {

      if(g_liConDeviceNumber_1 == DEMO_NO_LI_CONNECTED) {
        DemoLogf("Need to make a phone call.\n");        
      }
      else {
        // By enabling the Burst device, we start the process of transferring 
        // data from the hardware to the driver via DMA. The data is stored 
        // in the Burst device's buffer for later processing.
        DemoLogf("Enable g_hPhysDevs.hBurst\n");

        setupCaptureLiData();

        result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
        DemoCheckRc(g_hPhysDevs.hBurst, result);
      }
    }
  break;

  case 'N':
    {
      char sqzBuf[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;

      if(DemoIsBurstDeviceEnabled()) {
        nHighwayNo = 0;
        if(g_sBurstDevs.mBurstReceiver.bActive == OTX_FALSE || 
           g_sBurstDevs.mBurstReceiver.pRxFile == NULL) {
          
          // Always use TS 0
          g_sBurstDevs.mBurstReceiver.nTsToRecord = 0;
          
          g_sBurstDevs.mBurstReceiver.bActive = OTX_TRUE;

          DemoLogf("BurstReceiver[%d]. Please enter file to receive: ", 
                   nHighwayNo);
          OsaConsoleGetStr(sqzBuf);
          if(sqzBuf[0] == '\0') {
            DemoLogf("  No file enterred abort capture.\n");
            break;
          }
          

          g_sBurstDevs.mBurstReceiver.pRxFile = fopen(sqzBuf, "wb");
          if(g_sBurstDevs.mBurstReceiver.pRxFile == NULL) {
            DemoLogf("Error: Could not open file %s \n", sqzBuf);
            break;
          }
        } else {
          DemoLogf("Burst Receiver[%d] Thread already running.\n", nHighwayNo);
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
  break;

  case 'O':
    {
      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstReceiver.pRxFile) {
          DemoLogf("DemoBurstReceiver: Closing file\n");
          fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
          g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
        }
        g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
  break;

  // Use FXS 2 for audio playback. 
  case 'P':
    {
      OTX_INT32 nHwy;
      OTX_INT32 nTs;
      OTX_UINT8 sqTimeSlotMask[32];
      OTX_UINT32 slicState;

      DemoLogf("BurstSend[%d]. Enable Audio on FXS 2 \n");

      // Connect slic audio for device starting the call
      result = OtxSlicEnableAudio( g_hPhysDevs.hSlicLi[2], 
                                   &slicState );  
      DemoCheckRc(g_hPhysDevs.hSlicLi[2], result);
      // If burst device is not open, open it
      if(g_hPhysDevs.hBurst == OTX_INVALID_HANDLE_VALUE) {
        DemoLogf("Open Burst Device\n");

        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, 
                                          OTX_DEVICE_BURST, 
                                          0, 
                                          0, 
                                          g_hEventQueue, 
                                          &g_hPhysDevs.hBurst); 
        DemoCheckRc(g_hPhysDevs.hBoardController0, result);

        // set device buffer event trigger levels
        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  
                                               DEMO_BURST_BUF_SIZE); 
        DemoCheckRc(g_hPhysDevs.hBurst, result);
        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    
                                               DEMO_BURST_BUF_SIZE); 
        DemoCheckRc(g_hPhysDevs.hBurst, result);
      
      }

      // Check to see if it was setup perviously
      if(g_hPipeSendSlic == OTX_INVALID_HANDLE_VALUE) {
        DemoLogf("Connecting pipes\n");

        /*
        ** Burst send file
        */
        result = OtxDrvCreatePipe(64, &g_hPipeSendSlic);
        DemoCheckRc(g_hPipeSendSlic, result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
        nTs = 2 * 3; // calculate time slot
        nHwy = 6;    // SLIC Highway
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipeSendSlic, 
                                   OTX_PIPE_OUTPUT, 
                                   g_hPhysDevs.hTss,  
                                   nHwy, 
                                   sqTimeSlotMask, 
                                   32); 
        DemoCheckRc(g_hPipeSendSlic, result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
        nTs = 1;
        nHwy = 8;    // BURST highway
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipeSendSlic, 
                                   OTX_PIPE_INPUT, 
                                   g_hPhysDevs.hTss,  
                                   nHwy, 
                                   sqTimeSlotMask, 
                                   32); 
        DemoCheckRc(g_hPipeSendSlic, result);

        if(g_hPhysDevs.hBurst != OTX_INVALID_HANDLE_VALUE) {
          DemoLogf("Enable Burst Device\n");
          result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
          DemoCheckRc(g_hPhysDevs.hBurst, result);
        }
      } else {
        DemoLogf("The Burst send device is already enabled.\n");
      }
    }
  break;

  case '+':
    {
      // dump stats to debugger
      DemoLogf("Dump stats to debugger\n");
      OtxDrvDbgDump(g_hPhysDevs.hBurst, 100);
    }
    break;

  case 'V':
  {
    if (DemoIsBurstDeviceEnabled()) {
      g_sBurstDevs.mBurstReceiver.bVerbose = !g_sBurstDevs.mBurstReceiver.bVerbose;
      DemoLogf("Setting verbose setting to %s for Burst Receiver[%d] Thread\n", (g_sBurstDevs.mBurstReceiver.bVerbose)?"ON":"OFF", 0);
    } else {
      DemoLogf("The Burst device is not enabled. Command ignored\n");
    }
  }
  break;

  case 'p':
    printPhoneNumbers();
  break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
    // Select current Li
    {
      OTX_INT32 nLi = commandChar - '0';
      
      g_nCurrentLi = nLi;
      DemoLogf("SLIC #%d Selected ....\n", g_nCurrentLi);
    }
    break;


  break;
  case 'r':
    DemoLogf("Enter SLIC register (address) to be read : ");
    OsaConsoleGetStr(strBuf);

    regReadAddL = strtol(strBuf, NULL, 0);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            regReadAddL,
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("\n");

    DemoLogf("Register value was %d (%#X)\n", regValueL, regValueL);

  break;
  case 'w':
    DemoLogf("Enter SLIC register (address) to be written : ");
    OsaConsoleGetStr(strBuf);

    regReadAddL = strtol(strBuf, NULL, 0);

    printf("Enter value to be written ");
    OsaConsoleGetStr(strBuf);

    regValueL = strtol(strBuf, NULL, 0);
    regValue = (unsigned char) regValueL  & 0xFF;

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi], 
                             regReadAddL, 
                             regValue); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("Done\n");

  break;

  case 'R':
    DemoLogf("Enter FPGA register (address) to be read : ");
    OsaConsoleGetStr(strBuf);

    regReadAddL = strtol(strBuf, NULL, 0);

    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0,
                             regReadAddL,
                             &regValueL);
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    DemoLogf("\n");

    DemoLogf("Register value was %d (%#X)\n", regValueL, regValueL);

  break;
  case 'W':
    DemoLogf("Enter FPGA register (address) to be written : ");
    OsaConsoleGetStr(strBuf);

    regReadAddL = strtol(strBuf, NULL, 0);

    printf("Enter value to be written ");
    OsaConsoleGetStr(strBuf);

    regValueL = strtol(strBuf, NULL, 0);
    regValue = (unsigned char) regValueL  & 0xFF;

    result = OtxDrvWriteReg32(g_hPhysDevs.hBoardController0, 
                              regReadAddL, 
                              regValue); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    DemoLogf("Done\n");

  break;

  case 's':
    {
      OTX_UINT32 slicState;

      result = OtxSlicSendCallerID( g_hPhysDevs.hSlicLi[g_nCurrentLi],
                                    &slicState);
      DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

      result = OtxSlicStartRingState( g_hPhysDevs.hSlicLi[g_nCurrentLi], 
                                      &slicState );
      DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);  
    }
  break;

  case 'T':
    // This is only valid if the driver debug OtxDrv.sys is loaded
    // Display OTX Driver Current CID data
    result = OtxDrvDbgDump(g_hPhysDevs.hSlicLi[g_nCurrentLi], 400);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

  break;

  case 't':
    {
      OTX_UINT32 slicState;
      OTX_UINT32 slic;

      for(slic = 0; slic < DEMO_LI_MAX; slic++) {
        // set each slic to a different user ID
        sprintf(strBuf, "SLIC %d", slic);
        result = OtxSlicSetUserName( g_hPhysDevs.hSlicLi[slic],
                                     (OTX_UINT8 *) strBuf,
                                     &slicState);
        if(OTX_RESULT_CODE(result) != OTX_S_OK) {
          DemoLogf("\n  User ID for CID failed, SLIC state %d\n", 
                   slicState);
        }
        // set each slic to a different phone number
        sprintf(strBuf, "%d%d%d1234567", slic, slic, slic);

        result = OtxSlicSetPhoneNumber( g_hPhysDevs.hSlicLi[slic],
                                        (OTX_UINT8 *) strBuf,
                                        &slicState);
        if(OTX_RESULT_CODE(result) != OTX_S_OK) {
          DemoLogf("\n  Phone ID for CID failed, SLIC state %d\n", 
                   slicState);
        }
      }
    }
    DemoLogf("\nFinished CID setup\n");
  break;
  
  case 'u':
    {
      OTX_UINT32 slicState;

      DemoLogf("Enter User ID for CID for SLIC %d: ", g_nCurrentLi);
      OsaConsoleGetStr(strBuf);

      result = OtxSlicSetUserName( g_hPhysDevs.hSlicLi[g_nCurrentLi],
                                   (OTX_UINT8 *) strBuf,
                                   &slicState);
      if(OTX_RESULT_CODE(result) != OTX_S_OK) {
        DemoLogf("\n  User ID for CID failed, SLIC state %d\n", 
                 slicState);
      }

      DemoLogf("Enter 10 digit Phone ID for CID for SLIC %d: ", 
               g_nCurrentLi);
      OsaConsoleGetStr(strBuf);

      result = OtxSlicSetPhoneNumber( g_hPhysDevs.hSlicLi[g_nCurrentLi],
                                      (OTX_UINT8 *) strBuf,
                                      &slicState);
      if(OTX_RESULT_CODE(result) != OTX_S_OK) {
        DemoLogf("\n  Phone ID for CID failed, SLIC state %d\n", 
                 slicState);
      }
    }
    DemoLogf("\nFinished CID setup\n");

  break;

  case 'U':
    if(g_bSendCallerId == OTX_TRUE) {
      DemoLogf("Caller ID enabled\n");
    }
    else {
      DemoLogf("Caller ID disabled\n");
    }

    ringAllLines();

    DemoLogf("All phones should be ringing\n");

  break;

  case 'v':
    DemoLogf("Enter (y) to enable (n) to disable Caller ID : ");
    OsaConsoleGetStr(strBuf);

    if((strBuf[0] == 'y') || (strBuf[0] == 'Y')) {
      g_bSendCallerId = OTX_TRUE;      
    }
    else
    if((strBuf[0] == 'n') || (strBuf[0] == 'N')) {
      g_bSendCallerId = OTX_FALSE;      
    }

    if(g_bSendCallerId == OTX_TRUE) {
      DemoLogf("Caller ID enabled\n");
    }
    else {
      DemoLogf("Caller ID disabled\n");
    }
  break;

  case 'f':
    {
      OTX_BOOL bActive;
        
      DemoLogf("Toggling fail-over relay for SLIC #%d \n", g_nCurrentLi);
      
      result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hSlicLi[g_nCurrentLi], 
        OTX_ATTR_SLIC_RELAY_CONTROL, 
        &bActive); 
      
      if (bActive) {
        result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hSlicLi[g_nCurrentLi], 
          OTX_ATTR_SLIC_RELAY_CONTROL, 
          OTX_FALSE); 
        DemoLogf("Disengaged SLIC (pass-through state)\n");
      } else {
        result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hSlicLi[g_nCurrentLi], 
          OTX_ATTR_SLIC_RELAY_CONTROL, 
          OTX_TRUE); 
        DemoLogf("Engaged SLIC\n");
      }
    }
    break;
    
  case 'x':
    // This is only valid if the driver debug OtxDrv.sys is loaded
    // Display SLIC Direct Register
    result = OtxDrvDbgDump(g_hPhysDevs.hSlicLi[g_nCurrentLi], 200);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

  break;
  case 'z':
    // This is only valid if the driver debug OtxDrv.sys is loaded
    // Display SLIC Direct Register (Voltage and Current)
    result = OtxDrvDbgDump(g_hPhysDevs.hSlicLi[g_nCurrentLi], 300);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

  break;

  case '@':
    // Test SLIC using loop test
    DemoLogf("Enable SLIC loop test\n");

    g_RunNormalMode = OTX_FALSE;

  break;
    
  case '!':
    {
      DemoLogf("Terminating WatchDog Thread (Crash test)\n");
      g_bWatchDogThreadActive = OTX_FALSE;
    }
  break;


  case 'q':
  case 'Q':
    return 1;
  break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
  break;
  }

  return 0;
}



/*******************************************************************************

  @func

  Processes User Commands for SLIC audio path setup

  @comm

  This code is case sensitive.

*******************************************************************************/
OTX_INT32 DemoSlicReceiveConfigMenu(char commandChar)
{
  char strBuf[256];
  OTX_UINT32 regValueL = 0;
  OTX_RESULT result;

  switch(commandChar)   {
  case 'A':

    DemoLogf("Audio Transmit Path Gain\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 0 and 1 are used to set Analog receive path gain
    regValueL = regValueL & 0x0C;

    switch(regValueL)   {

      case 0:
        DemoLogf("Gain set to 0 dB\n");
      break;
      case 0x04:
        DemoLogf("Gain set to -3.5 dB\n");
      break;
      case 0x08:
        DemoLogf("Gain set to 3.5 dB\n");
      break;
      case 0x0C:
        DemoLogf("Analog transmit path muted\n");
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    DemoLogf("Enter Option \n");
    DemoLogf("0: for O db Gain\n");
    DemoLogf("1: for -3.5 db Gain\n");
    DemoLogf("2: for 3.5 db Gain\n");
    DemoLogf("3: Analog receive path muted\n");
    DemoLogf("> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);
    regValueL = regValueL & ~(0x0C);

    switch(strBuf[0])   {

      case '0':
        DemoLogf("Gain set to 0 dB\n");
      break;
      case '1':
        DemoLogf("Gain set to -3.5 dB\n");
        regValueL = regValueL | 0x04;
      break;
      case '2':
        DemoLogf("Gain set to 3.5 dB\n");
        regValueL = regValueL | 0x08;
      break;
      case '3':
        DemoLogf("Analog transmit path muted\n");
        regValueL = regValueL | 0x0C;
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                             9,     // Audio gain control register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);


    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;

  case 'a':

    DemoLogf("Audio Receive Path Gain\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 0 and 1 are used to set Analog receive path gain
    regValueL = regValueL & 0x03;

    switch(regValueL)   {

      case 0:
        DemoLogf("Gain set to 0 dB\n");
      break;
      case 1:
        DemoLogf("Gain set to -3.5 dB\n");
      break;
      case 2:
        DemoLogf("Gain set to 3.5 dB\n");
      break;
      case 3:
        DemoLogf("Analog receive path muted\n");
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    DemoLogf("Enter Option \n");
    DemoLogf("0: for O db Gain\n");
    DemoLogf("1: for -3.5 db Gain\n");
    DemoLogf("2: for 3.5 db Gain\n");
    DemoLogf("3: Analog receive path muted\n");
    DemoLogf("> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);
    regValueL = regValueL & ~(0x03);

    switch(strBuf[0])   {

      case '0':
        DemoLogf("Gain set to 0 dB\n");
      break;
      case '1':
        DemoLogf("Gain set to -3.5 dB\n");
        regValueL = regValueL | 0x01;
      break;
      case '2':
        DemoLogf("Gain set to 3.5 dB\n");
        regValueL = regValueL | 0x02;
      break;
      case '3':
        DemoLogf("Analog receive path muted\n");
        regValueL = regValueL | 0x03;
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                             9,     // Audio gain control register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);


    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;

  case 'b':

    DemoLogf("Audio Hybrid Gain\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            11,     // Audio Hybrid gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 0, 1, and 2 are used to set Hybrid gain
    regValueL = regValueL & 0x07;

    switch(regValueL)   {

      case 0:
        DemoLogf("Gain set to 4.08 dB\n");
      break;
      case 1:
        DemoLogf("Gain set to 2.5 dB\n");
      break;
      case 2:
        DemoLogf("Gain set to 1.16 dB\n");
      break;
      case 3:
        DemoLogf("Gain set to 0 dB\n");
      break;
      case 4:
        DemoLogf("Gain set to -1.02 dB\n");
      break;
      case 5:
        DemoLogf("Gain set to -1.94 dB\n");
      break;
      case 6:
        DemoLogf("Gain set to -2.77 dB\n");
      break;
      case 7:
        DemoLogf("Audio Hybrid off\n");
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    DemoLogf("Enter Option \n");
    DemoLogf("0: for 4.08 db Gain\n");
    DemoLogf("1: for 2.5 db Gain\n");
    DemoLogf("2: for 1.16 db Gain\n");
    DemoLogf("3: for O db Gain\n");
    DemoLogf("4: for -1.02 db Gain\n");
    DemoLogf("5: for -1.94 db Gain\n");
    DemoLogf("6: for -2.77 db Gain\n");
    DemoLogf("7: Audio Hybrid off\n");
    DemoLogf("> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            11,     // Audio Hybrid gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    regValueL = regValueL & ~(0x07);

    switch(strBuf[0])   {

      case '0':
        DemoLogf("Gain set to 4.08 dB\n");
      break;
      case '1':
        DemoLogf("Gain set to 2.5 dB\n");
        regValueL = regValueL | 0x01;
      break;
      case '2':
        DemoLogf("Gain set to 1.16 dB\n");
        regValueL = regValueL | 0x02;
      break;
      case '3':
        DemoLogf("Gain set to 0 dB\n");
        regValueL = regValueL | 0x03;
      break;
      case '4':
        DemoLogf("Gain set to -1.02 dB\n");
        regValueL = regValueL | 0x04;
      break;
      case '5':
        DemoLogf("Gain set to -1.94 dB\n");
        regValueL = regValueL | 0x05;
      break;
      case '6':
        DemoLogf("Gain set to -2.77 dB\n");
        regValueL = regValueL | 0x06;
      break;
      case '7':
        DemoLogf("Audio Hybrid off\n");
        regValueL = regValueL | 0x07;
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                             11,     // Audio Hybrid gain control register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);
    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            11,     // Audio Hybrid gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;

  case 'c':

    DemoLogf("DC-DC Converter Voice Band Noise Squelch\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 1 used to enable audio band noise squelch
    regValueL = regValueL & 0x02;

    if(regValueL) {
      DemoLogf("Voice band squelch enabled\n");
    }
    else {
      DemoLogf("Voice band squelch disabled\n");
    }

    DemoLogf("Enter (y) to enable (n) to disable squelch \n> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);


    if((strBuf[0] == 'y') || (strBuf[0] == 'Y')) {
      regValueL = regValueL | 0x02;
    }
    else
    if((strBuf[0] == 'n') || (strBuf[0] == 'N')) {
      regValueL = regValueL & ~(0x02);
    }
    else {
      DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                             108,     // Si3210 Enhancement register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);
    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;


  case 'd':

    DemoLogf("DC-DC Converter Voice Band Noise Squelch Responce Time\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 1 used to enable audio band noise squelch
    regValueL = regValueL & 0x20;

    if(regValueL) {
      DemoLogf("Voice band squelch ENHANCED Algorithm\n");
    }
    else {
      DemoLogf("Voice band squelch NORMAL Algorithm\n");
    }

    DemoLogf("Enter (y) to enhanced (n) to Normal squelch \n> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);


    if((strBuf[0] == 'y') || (strBuf[0] == 'Y')) {
      regValueL = regValueL | 0x20;
    }
    else
    if((strBuf[0] == 'n') || (strBuf[0] == 'N')) {
      regValueL = regValueL & ~(0x20);
    }
    else {
      DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                             108,     // Si3210 Enhancement register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);
    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_nCurrentLi],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_nCurrentLi], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;


  case 'Q':
    return 1;
  break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
    // Select current Li
    {
      OTX_INT32 nLi = commandChar - '0';
      
      g_nCurrentLi = nLi;
      DemoLogf("SLIC #%d Selected ....\n", g_nCurrentLi);
    }
  break;

  default:
    DemoLogf("\nCmd Si3210 Audio Receive Path Configuration\n");
    DemoLogf("===  ========\n");
    DemoLogf(" Q   Return to Main Menu\n");
    DemoLogf(" 0-5 Select active SLIC. #%d is now active\n", g_nCurrentLi);
    DemoLogf(" A   Set Transmit Audio Gain\n");
    DemoLogf(" a   Set Receive Audio Gain\n");
    DemoLogf(" b   Set Receive Hybrid Gain\n");
    DemoLogf(" c   Set DC-DC Converter Noise Squelch\n");
    DemoLogf(" d   Set DC-DC Converter Noise Squelch Response\n");
    DemoLogf("\n > ");
  break;
  }
  
  return 0;
}


