/* $Workfile: OtxDsp.h $   $Revision: 28 $    $Modtime: 8/05/08 5:33p $  */
#ifndef OTX_DSP_H
#define OTX_DSP_H
/*******************************************************************************

  @doc PublicHwApi OtxDsp

  Copyright (C) 1998 - 2003 Odin TeleSystems Inc., All Rights Reserved

  OTX Generic Hardware Driver API

  @module
  OtxDsp | This file contains prototypes for the DSP Management Functions

  @contents2
  OtxDsp.h |

  @index func | OtxDsp

  @contents2
  OTX DSP Device Management Functions - OtxDspXXX() | The user functions available for
  the Management of DSP Devices

  @head3 OtxDspXXX() Function Reference |

  @index func | OtxDsp

*******************************************************************************/

#include "OtxType.h"    // OTX specific type definitions
#include "OtxDef.h"     // OTX specific definitions

#include "OtxDspD.h"    // OTX DSP type definitions

#ifdef __cplusplus
extern "C" {
#endif



/*******************************************************************************

  @func
    Load and Run a program on a DSP from a COFF format (*.out) file

  @ex |

    // open physical device of DSP 
    for (i = 0; i < 2; i++) {
      result = OtxDrvOpenPhysicalDevice(
                  g_hPhysDevs.hBoardController, // Vidar series or Thor-2-PCMCIA series 
                  OTX_DEVICE_DSP, 
                  i, 
                  i, 
                  g_hEventQueue, 
                  &g_hPhysDevs.hDsp[i]); 
    }
    
    // load DSP program from file OtxSpm1.out
    for (nDsp = 0; nDsp < 2; nDsp++) {
      result = OtxDspRunProgram(
                  g_hPhysDevs.hDsp[nDsp], 
                  "OtxSpm1.out", 
                  "_c_int00"); 
    }

  @xref
    <f OtxDspRunSpm>
    <f OtxDspRestartProgram>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspRunProgram(
  IN OTX_HANDLE     hDspDevice,         // @parm Handle to DSP device
  IN const char     *szPrgfile,         // @parm Application file to be loaded
  IN const char     *szEntryPoint       // @parm Label for the entry point of the application
);


/*******************************************************************************

  @func
    Load and Run an OTX Standard Program Module (SPM) on a DSP from the a 
    program image stored in the OTX lib.
  
  @ex |
  
    // open physical device of DSP 
    for (i = 0; i < 2; i++) {
      result = OtxDrvOpenPhysicalDevice(
                  g_hPhysDevs.hBoardController, // Vidar series or Thor-2-PCMCIA series 
                  OTX_DEVICE_DSP, 
                  i, 
                  i, 
                  g_hEventQueue, 
                  &g_hPhysDevs.hDsp[i]); 
    }
  
  
    // load DSP Program OTX_DSP_SPM_OTXSPM1
    for (i = 0; i < 2; i++) {
      result = OtxDspRunSpm(
                  g_hPhysDevs.hDsp[i], 
                  OTX_DSP_SPM_OTXSPM1); 
    }

  @xref
    <f OtxDspRunProgram>
    <t OtxDspSpmE>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspRunSpm(
  IN OTX_HANDLE     hDspDevice,         // @parm Handle to DSP device
  IN OtxDspSpmE     eProgramModule      // @parm Standard Program Module to be loaded in the DSP (from the Lib)
);


/*******************************************************************************

  @func
    Convert a DSP Application file from COFF format (*.out) for a C548/549 DSP 
    to the OTX DSP Image format (used by <f OtxDspRunImage>). 
  
  @ex |

    OTX_UINT16 *psqProgImage;
    OTX_UINT16 coProgImage;
    OTX_INT32  nEntryAddr;
    const OTX_UINT16 DEMO_DSP_IMAGE_BUF_SIZE = 60000;
    
    psqProgImage = (OTX_UINT16 *)malloc(DEMO_DSP_IMAGE_BUF_SIZE*sizeof(OTX_UINT16));
    
    // convert "OtxSpm1.out" to DSP image format
    result = OtxDspCreateProgramImage(
                "OtxSpm1.out", 
                "_c_int00", 
                DEMO_DSP_IMAGE_BUF_SIZE, 
                psqProgImage, 
                &coProgImage, 
                &nEntryAddr); 

    // load DSP program with image format
    for (nDsp = 0; nDsp < 2; nDsp++) {
      result = OtxDspRunImage(
                  g_hPhysDevs.hDsp[nDsp], 
                  psqProgImage, 
                  coProgImage, 
                  nEntryAddr); 
    }
    
    free(psqProgImage);

  @xref
    <f OtxDspRunImage>
    <f OtxDspRunProgram>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspCreateProgramImage(
  IN const char     *szProgfile,        // @parm Application file to be converted to an image   
  IN const char     *szEntryPoint,      // @parm Label for the entry point of the application. Usually "_c_int00" for a C-compiled application.
  IN OTX_UINT16     coProgImageMax,     // @parm Count of objects (size) of psqProgImage (in elements) provided by the user
  IN OUT OTX_UINT16 *psqProgImage,      // @parm Image Buffer provided by the user
  OUT OTX_UINT16    *coProgImage,       // @parm Number of image elements returned in psqProgImage 
  OUT OTX_INT32     *nEntryAddr         // @parm Program Entry address
);


/*******************************************************************************

  @func
    Converts a DSP Application file from COFF format (*.out) for C5400 and 
    C5510 DSPs to the OTX DSP Extended Image format (used by <f OtxDspRunImage>).

  @ex |
    
    #define IMAGE_BUF_SIZE 128000
    static OTX_UINT16 g_psqProgImage[IMAGE_BUF_SIZE];

    // convert "OtxSpm5.out" to DSP Extended Image format
    result = OtxDspCreateProgramImageEx(
                "OtxSpm5.out", 
                "_c_int00", 
                IMAGE_BUF_SIZE, 
                g_psqProgImage, 
                &coProgImage, 
                &nEntryAddr);
    
    // load DSP program with image format
    for (nDsp = 0; nDsp < 2; nDsp++) {
      result = OtxDspRunImage(
                  g_hPhysDevs.hDsp[nDsp], 
                  psqProgImage, 
                  coProgImage, 
                  nEntryAddr); 
    }
    
    free(psqProgImage);

  @xref
    <f OtxDspRunImage>
    <f OtxDspRunProgram>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspCreateProgramImageEx(
  IN const char     *szProgfile,        // @parm Application file to be converted to an image   
  IN const char     *szEntryPoint,      // @parm Label for the entry point of the application. Usually "_c_int00" for a C-compiled application.
  IN OTX_UINT32     coProgImageMax,     // @parm Count of objects (size) of psqProgImage (in elements) provided by the user
  IN OUT OTX_UINT16 *psqProgImage,      // @parm Image Buffer provided by the user
  OUT OTX_UINT32    *coProgImage,       // @parm Number of image elements returned in psqProgImage 
  OUT OTX_INT32     *nEntryAddr         // @parm Program Entry address
  );


/*******************************************************************************

  @func
    Load and start a DSP application program. The DSP application program must 
    be in the OTX DSP Image format (created by a call 
    to <f OtxDspCreateProgramImage>). 

  @ex |
    see example on       
    @ex <f OtxDspCreateProgramImage> |
    @ex <f OtxDspCreateProgramImageEx> |

  @xref
    <f OtxDspCreateProgramImage>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspRunImage(
  IN OTX_HANDLE     hDspDevice,         // @parm Handle to DSP device
  IN OTX_UINT16     *psqProgImage,      // @parm Program image 
  IN OTX_UINT16     coProgImage,        // @parm Number of elements in psqProgImage 
  IN OTX_INT32      nEntryAddr          // @parm Program Entry Address 
);


/*******************************************************************************

  @func
    Load and start a DSP application program. The DSP application program must 
    be in the OTX DSP Image format (created by a call 
    to <f OtxDspCreateProgramImage>). 

    Number of elements is 32-bit (OTX_INT32) which is different from 
    <f OtxDspRunImage> 

  @ex |
    see example on       
    @ex <f OtxDspCreateProgramImageEx> |

  @xref
    <f OtxDspCreateProgramImageEx>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspRunImageEx(
  IN OTX_HANDLE     hDspDevice,         // @parm Handle to DSP device
  IN OTX_UINT16     *psqProgImage,      // @parm Program image 
  IN OTX_UINT32     coProgImage,        // @parm Number of elements in psqProgImage (32-bit) 
  IN OTX_INT32      nEntryAddr          // @parm Program Entry Address 
);


/*******************************************************************************

  @func
    DSP Debugging utility function. Restarts a previously loaded and/or 
    running program. 
    
    This function is only meant to be called if a DSP Debugger is being
    used. 

  @ex This function should only be used on the following call sequence: |

    1. First the OtxDspRunImage(), OtxDspRunSpm(), or OtxDspRunProgram() is called.
    2. Immediately thereafter the host application must be stopped.
    3. While the host application is stopped the DSP program is loaded and started 
       from the DSP Debugger.
    4. Once the DSP program is running again, OtxDspRunProgram() should be called
       to bring the DSP program back to the state it was right after step 1 was
       executed.

  @ex |

    // load DSP program
    for (nDsp = 0; nDsp < 2; nDsp++) {
      result = OtxDspRunSpm(
                  g_hPhysDevs.hDsp[nDsp], 
                  OTX_DSP_SPM_OTXSPM1); 
    }

    // load and run DSP program to all of the DSPs in the DSP Debugger

    // restart program on DSPs
    for (i = 0; i < 2; i++) {
        result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); 
    }

  @xref
    <f OtxDspRunProgram>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspRestartProgram(
  IN OTX_HANDLE     hDspDevice          // @parm Handle to DSP device
);


/*******************************************************************************

  @func
    Send a user command to a logical <e OtxLogicalDeviceTypeE.OTX_LDEVICE_USER_APPLICATION> 
    device (hosted by a DSP or DMP device).

  @ex |

      Demo1IoInSetBlinkFreqS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo1Rc;
      
      sIn.nBlinkCountFactor = DEMO_FAST_BLINK_COUNT_FACTOR;
      sIn.nDummyData = 6;
      
      // sending Blink Count factor DEMO_FAST_BLINK_COUNT_FACTOR to User App 
      result = OtxDspIoControl(
                  g_hLogicDevs.hUserApp, 
                  DEMO1_IOCTL_SET_BLINK_FREQ, 
                  &sIn, 
                  sizeof(sIn), 
                  &nDemo1Rc, 
                  sizeof(nDemo1Rc), 
                  &nBytesReturned, 
                  OTX_TASK_SYNC, 
                  OTX_TIME_INFINITY); 

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspIoControl(
  IN  OTX_HANDLE    hDevice,            // @parm Handle to a <e OtxLogicalDeviceTypeE.OTX_LDEVICE_USER_APPLICATION> logical device
  IN  OTX_UINT16    nIoControlCode,     // @parm I/O Control Code to be passed 
  IN  void          *pInBuf,            // @parm Pointer to the user allocated input buffer
  IN  OTX_UINT32    nInBufSize,         // @parm Size of the input buffer
  IN  void          *pOutBuf,           // @parm Pointer to the user allocated output buffer
  IN  OTX_UINT32    nOutBufSize,        // @parm Size of the output buffer
  OUT OTX_UINT32    *pnBytesReturned,   // @parm Number of Bytes returned
  IN  OTX_TASK_REF  nTaskRef,           // @parm See <t OTX_TASK_REF>
  IN  OTX_TIME      nWaitMaxMs          // @parm See <t OTX_TIME>
);


/*******************************************************************************

  @func
    Read data from a user supplied DSP application

  @comm
    Usually it is used in event case OTX_LDEVICE_USER_APPLICATION of user 
    DSP application

  @rdesc Returns one of the following values:

    @flag OTX_S_OK | Data was fetched
    @flag OTX_S_FALSE | No Data available at this moment
    @flag OTX_E_BUFFER_TO_SMALL | The supplied user data buffer is too small to fit the received data

  @ex |
    // Note: function DemoLogf() is used to display message in our demo programs 

    case OTX_LDEVICE_USER_APPLICATION:
      if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
        OTX_UINT8  buf[8192];
        OTX_UINT32 nBytesReturned;
        OTX_UINT32 nIoControlCode;

        otxRc = OtxDspReadData(
                   pEventData->m_hDevice, 
                   OTX_TASK_SYNC, 
                   buf, 
                   8192, 
                   &nBytesReturned, 
                   &nIoControlCode);
        
        if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
          switch(nIoControlCode) {
            case DEMO1_NOTIFY_DEBUG_DATA:
            {
              Demo1NotifyGetDebugDataS *pDataS = (Demo1NotifyGetDebugDataS *)buf;
              if (nBytesReturned != sizeof(Demo1NotifyGetDebugDataS)) {
                DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
              } else {
                DemoLogf("DEMO1_NOTIFY_DEBUG_DATA (nRequestId=%d, nBytes=%d)\n", pEventData->m_nRequestId, nBytesReturned);
                DemoLogf("nData0 = %.4X\n", pDataS->nData0);
                DemoLogf("nData1 = %.4X\n", pDataS->nData1);
                DemoLogf("nData2 = %.4X\n", pDataS->nData2);

                DemoLogf("sqBuf[299] = %.4X\n", pDataS->sqBuf[299]);
              }
            }
            break;
        
            default:
              DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
            break;
          }
      
        } else {
        
          OTX_RESULT errorCode = otxRc;
          otxRc = OtxDrvResultCode2String(
                     pEventData->m_hDevice, 
                     errorCode, 
                     256, 
                     &msgStr[0]);
          
          if (OTX_SUCCESS(otxRc)) {
            DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
          } else {
            DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
          }
        }

      } else {

        otxRc = OtxDrvEventCode2String(
                   pEventData->m_hDevice, 
                   pEventData->m_nCode, 
                   256, 
                   &msgStr[0]);

        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
        } else {
          DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
        }
      }
    break;  // end of OTX_LDEVICE_USER_APPLICATION

  @xref

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspReadData(
  IN  OTX_HANDLE    hDevice,            // @parm Handle to a <e OtxLogicalDeviceTypeE.OTX_LDEVICE_USER_APPLICATION> logical device
  IN  OTX_TASK_REF  nTaskRef,           // @parm See <t OTX_TASK_REF>
  OUT OTX_UINT8     *pOutBuf,           // @parm Pointer to the user allocated output buffer
  IN  OTX_UINT32    nOutBufSize,        // @parm Size of the output buffer
  OUT OTX_UINT32    *pnBytesReturned,   // @parm Number of Bytes returned
  OUT OTX_UINT32    *pnIoControlCode    // @parm I/O Control Code the returned data is associated with
);


/*******************************************************************************

  @func
    Read a number of octets from the DSP's Host Port Interface (HPI)
    into a user supplied buffer.

  @comm
    Note: This function should be used with caution. It is intended for
    debugging purposes and advanced users only.

    API <f OtxDspWriteHpi> can be used to write the data  

  @ex |
  
    OTX_UINT32 i;
    OTX_UINT16 sqReadBuf[0x800];

    // dump the entire HPI memory for the current DSP
    result = OtxDspReadHpi(
                g_hPhysDevs.hDsp, 
                0x1000, 
                0x800, 
                sqReadBuf); 

    // display the data
    for (i = 0; i < 0x800; i++) {
        if (i && ((i % 8) == 0)) 
        DemoLogf("\n %.4X ", sqReadBuf[i]);
      }
    }


  @xref
    <f OtxDspWriteHpi>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspReadHpi(
  IN  OTX_HANDLE    hDspDevice,         // @parm Handle to the physical DSP device
  IN  OTX_UINT32    nHpiAddress,        // @parm HPI Address to start to read from 
  IN  OTX_UINT32    coWordsToRead,      // @parm Number of OTX_UINT16s (Words) to read from the HPI
  IN OUT OTX_UINT16 *pReadBuf           // @parm Pointer to the user allocated output buffer (Words)
);


/*******************************************************************************

  @func
    Write a number of octets to the DSP's Host Port Interface (HPI)
    from a user supplied buffer.

  @comm
    Note: This function should be used with caution since it can corrupt the
    communication buffers between the DSP and the host. It is intended for
    debugging purposes and advanced users only.

    API <f OtxDspReadHpi> can be used to check the data written  

  @ex |
  
    static OTX_UINT16 writeVal = 0x5A5A;
    OTX_UINT16 sqWriteBuf[1];

    sqWriteBuf[0] = writeVal;

    // write a value to the last address in the HPI memory for the current DSP
    result = OtxDspWriteHpi(
                g_hPhysDevs.hDsp, 
                0x17FF, 
                1, 
                sqWriteBuf); 

  @xref
    <f OtxDspReadHpi>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDspWriteHpi(
  IN  OTX_HANDLE    hDspDevice,         // @parm Handle to the physical DSP device
  IN  OTX_UINT32    nHpiAddress,        // @parm HPI Address to start to write to
  IN  OTX_UINT32    coWordsToWrite,     // @parm Number of OTX_UINT16s (Words) to write to the HPI
  IN OUT OTX_UINT16 *pWriteBuf          // @parm Pointer to the user allocated input buffer (Words)
);



#ifdef __cplusplus
}  // extern "C"
#endif

/*******************************************************************************

  Copyright (C) Odin TeleSystems Inc., 1998 - 2003

  All rights reserved. No part of this computer program may be reproduced,
  stored in any retrieval system, or transmitted, in any form or by any
  means, electronic, mechanical, photocopying, recording, or otherwise,
  without the prior written permission of Odin TeleSystems Inc., or without
  a permission stipulated in the agreement/contract under which this computer
  program was supplied.

*******************************************************************************/

#endif    // OTX_DSP_H
